#!/bin/env python


"""
RetryAlgoBase

This is the base class for Retry Algos
"""

from builtins import object
import time
import datetime
import logging

class RetryAlgoBase(object):
    """
    A RetryAlgo plugin for jobs in the Create state

    """

    def __init__(self, config):
        object.__init__(self)
        self.config = config

    def setup(self, config):
        """
        Pass in config (WMFactory too stupid to do so on init)

        """
        self.config = config

    def isReady(self, job, cooloffType):
        """
        Actual function that does the work

        """

        pass

    def convertdatetime(self, t):
        return int(time.mktime(t.timetuple()))

    def timestamp(self):
        """
        generate a timestamp
        """
        t = datetime.datetime.now()
        return self.convertdatetime(t)

    def getAlgoParam(self, jobType, param = 'coolOffTime', defaultReturn = {}):
        """
        _getAlgoParam_

        Get a parameter from the config for the current algorithm and given
        job type
        """
        pluginName = self.__class__.__name__
        pluginArgs = getattr(self.config.RetryManager, pluginName)

        if hasattr(pluginArgs, jobType):
            algoParams = getattr(pluginArgs, jobType)
        else:
            algoParams = pluginArgs.default

        if hasattr(algoParams, param):
            return getattr(algoParams, param)
        else:
            logging.error("No %s for %s algorithm and %s job type" % (param, pluginName, jobType))
            return defaultReturn



#!/usr/bin/env python
"""
_CompleteInput_

Oracle implementation of Jobs.CompleteInput
"""

from WMCore.WMBS.MySQL.Jobs.CompleteInput import CompleteInput as MySQLCompleteInput

class CompleteInput(MySQLCompleteInput):
    fileSelect = """SELECT job_files.subscriptionid, job_files.fileid, job_files.lfn,
                           job_files.jobid,
                           COUNT(wmbs_job_assoc.job) AS total, SUM(wmbs_job.outcome) AS numsuccess FROM
                      (SELECT wmbs_jobgroup.subscription AS subscriptionid,
                              wmbs_job_assoc.job AS jobid,
                              wmbs_job_assoc.fileid AS fileid,
                              wmbs_file_details.lfn AS lfn FROM wmbs_job_assoc
                         INNER JOIN wmbs_job ON
                           wmbs_job_assoc.job = wmbs_job.id
                         INNER JOIN wmbs_jobgroup ON
                           wmbs_job.jobgroup = wmbs_jobgroup.id
                         INNER JOIN wmbs_file_details ON
                           wmbs_job_assoc.fileid = wmbs_file_details.id
                       WHERE wmbs_job.id = :jobid) job_files
                      INNER JOIN wmbs_job_assoc ON
                        job_files.fileid = wmbs_job_assoc.fileid
                      INNER JOIN wmbs_job ON
                        wmbs_job_assoc.job = wmbs_job.id
                      INNER JOIN wmbs_jobgroup ON
                        wmbs_job.jobgroup = wmbs_jobgroup.id
                    WHERE wmbs_jobgroup.subscription = job_files.subscriptionid
                    GROUP BY job_files.subscriptionid, job_files.fileid, job_files.lfn,
                             job_files.jobid
                    HAVING COUNT(wmbs_job_assoc.job) = SUM(wmbs_job.outcome)"""

    sql = """INSERT INTO wmbs_sub_files_complete (fileid, subscription)
               SELECT :fileid, :subid FROM DUAL
               WHERE NOT EXISTS
                 (SELECT * FROM wmbs_sub_files_complete
                  WHERE fileid = :fileid AND subscription = :subid)"""

    failSql = """INSERT INTO wmbs_sub_files_failed (fileid, subscription)
                   SELECT :fileid, :subid FROM DUAL
                   WHERE NOT EXISTS
                     (SELECT * FROM wmbs_sub_files_failed
                      WHERE fileid = :fileid AND subscription = :subid)"""



#!/usr/bin/env python
"""
_LoadFromUID_

Oracle implementation of JobGroup.LoadFromUID
"""

__all__ = []



from WMCore.WMBS.MySQL.JobGroup.LoadFromUID import LoadFromUID as LoadFromUIDMySQL

class LoadFromUID(LoadFromUIDMySQL):
    pass



#!/usr/bin/env python
"""
_harvestingInjector_

"""
from __future__ import print_function

import os
import sys
import threading
import time
from argparse import ArgumentParser

from DBSAPI.dbsApi import DbsApi
from WMCore.WMSpec.StdSpecs.Harvesting import harvestingWorkload, getTestArguments

from WMCore.DataStructs.Run import Run
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMInit import connectToDB
from WMCore.WMSpec.Makers.TaskMaker import TaskMaker
from WMCore.WorkQueue.WMBSHelper import WMBSHelper


def check_list(option, opt, value):
    return value.split(",")


def comma_separated_list(string):
    return string.split(',')


usage = "usage: %prog [options]"
parser = ArgumentParser(usage=usage)
parser.add_argument("-d", "--dataset", dest="InputDataset",
                    action="store", help="Dataset to harvest",
                    metavar="DATASET")
parser.add_argument("-R", "--run", dest="RunWhitelist", type=comma_separated_list,
                    action="store", help="Comma separated list of runs",
                    metavar="RUN1,RUN2", default=[])
parser.add_argument("-r", "--release", dest="CMSSWVersion",
                    action="store", help="CMSSW version to use for harvesting",
                    metavar="CMSSW_X_Y_Z")
parser.add_argument("-s", "--scenario", dest="Scenario",
                    action="store", help="Configuration/DataProcessing scenario",
                    metavar="SCENARIO")
parser.add_argument("-t", "--global-tag", dest="GlobalTag",
                    action="store", help="Conditions global tag",
                    metavar="GLOBALTAG")
parser.add_argument("-f", "--reference", dest="RefHistogram",
                    action="store", help="Reference histogram",
                    metavar="LFN")

options = parser.parse_args()

missing = []
mandatory = ["InputDataset", "CMSSWVersion", "Scenario", "GlobalTag"]
for option in options.__dict__:
    if getattr(options, option) is None and option in mandatory:
        missing.append(option)
if missing:
    print("Error: The following mandatory options are missing:")
    print("\n".join(missing))
    sys.exit(1)

# The default arguments are set in:
#   WMCORE/src/python/WMCore/WMSpec/StdSpecs/Harvesting.py
arguments = getTestArguments()
arguments.update(options.__dict__)

connectToDB()

req_time = "%.2f" % time.time()
workloadName = "Harvesting%s--%s" % (arguments["InputDataset"].replace("/", "__"), req_time)
workloadFile = "Harvesting%s--%s.pkl" % (arguments["InputDataset"].replace("/", "__"), req_time)
os.mkdir(workloadName)
workload = harvestingWorkload(workloadName, arguments)
workloadPath = os.path.join(workloadName, workloadFile)
workload.setOwner("sfoulkes@fnal.gov")
workload.setSpecUrl(workloadPath)

# Build a sandbox using TaskMaker
taskMaker = TaskMaker(workload, os.path.join(os.getcwd(), workloadName))
taskMaker.skipSubscription = True
taskMaker.processWorkload()

workload.save(workloadPath)


def injectFilesFromDBS(inputFileset, datasetPath, runsWhiteList=[]):
    """
    _injectFilesFromDBS_

    """
    print("injecting files from %s into %s, please wait..." % (datasetPath, inputFileset.name))
    args = {}
    args["url"] = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader"
    args["version"] = "DBS_2_1_1"
    args["mode"] = "GET"
    dbsApi = DbsApi(args)
    dbsResults = dbsApi.listFileArray(path=datasetPath, retriveList=["retrive_lumi", "retrive_run"])
    print("  found %d files, inserting into wmbs..." % (len(dbsResults)))

    for dbsResult in dbsResults:
        if runsWhiteList and str(dbsResult["LumiList"][0]["RunNumber"]) not in runsWhiteList:
            continue
        myFile = File(lfn=dbsResult["LogicalFileName"], size=dbsResult["FileSize"],
                      events=dbsResult["NumberOfEvents"], checksums={"cksum": dbsResult["Checksum"]},
                      locations="cmssrm.fnal.gov", merged=True)
        myRun = Run(runNumber=dbsResult["LumiList"][0]["RunNumber"])
        for lumi in dbsResult["LumiList"]:
            myRun.appendLumi(lumi["LumiSectionNumber"])
        myFile.addRun(myRun)
        myFile.create()
        inputFileset.addFile(myFile)

    if len(inputFileset) < 1:
        raise Exception("No files were selected!")

    inputFileset.commit()
    inputFileset.markOpen(False)
    return


myThread = threading.currentThread()
myThread.transaction.begin()
for workloadTask in workload.taskIterator():
    inputFileset = Fileset(name=workloadTask.getPathName())
    inputFileset.create()

    inputDataset = workloadTask.inputDataset()
    inputDatasetPath = "/%s/%s/%s" % (inputDataset.primary,
                                      inputDataset.processed,
                                      inputDataset.tier)
    injectFilesFromDBS(inputFileset, inputDatasetPath, options.RunWhitelist)

    myWMBSHelper = WMBSHelper(workload)
    myWMBSHelper._createSubscriptionsInWMBS(workloadTash.getPathName())

myThread.transaction.commit()



"""
WorkQueueElement

A dictionary based object meant to represent a WorkQueue element
"""

from future.utils import viewitems, viewvalues
import time
from hashlib import md5

from Utils.Utilities import encodeUnicodeToBytesConditional
from Utils.PythonVersion import PY3

STATES = ('Available', 'Negotiating', 'Acquired', 'Running',
          'Done', 'Failed', 'CancelRequested', 'Canceled')


def possibleSites(element):
    """
    _possibleSites_

    Checks the site and data restrictions and return a list of possible sites
    to work on this element.
    """
    # check if the whole document was provide
    elem = element.get('WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement', element)

    if elem['NoInputUpdate'] and elem['NoPileupUpdate']:
        return elem['SiteWhitelist']

    commonSites = set(elem['SiteWhitelist']) - set(elem['SiteBlacklist'])

    if elem['Inputs'] and elem['NoInputUpdate'] is False:
        commonSites = commonSites.intersection(set([y for x in viewvalues(elem['Inputs']) for y in x]))
    if elem['ParentFlag'] and elem['NoInputUpdate'] is False:
        commonSites = commonSites.intersection(set([y for x in viewvalues(elem['ParentData']) for y in x]))
    if elem['PileupData'] and elem['NoPileupUpdate'] is False:
        commonSites = commonSites.intersection(set([y for x in viewvalues(elem['PileupData']) for y in x]))

    return list(commonSites)


class WorkQueueElement(dict):
    """Class to represent a WorkQueue element"""

    def __init__(self, **kwargs):
        dict.__init__(self)

        if 'Status' in kwargs and kwargs['Status'] not in STATES:
            msg = 'Invalid Status: %s' % kwargs['Status']
            raise ValueError(msg)

        self.update(kwargs)

        self._id = None

        # XXX If adding or modifying any new parameter which affects the
        # XXX workflow or data run over, the id function must be updated

        self.setdefault('Inputs', {})
        self.setdefault('ProcessedInputs', [])
        self.setdefault('RejectedInputs', [])
        # Some workflows require additional datasets for PileUp
        # Track their locations
        self.setdefault('PileupData', {})
        # both ParentData and ParentFlag is needed in case there Dataset split,
        # even though ParentFlag is True it will have empty ParentData
        self.setdefault('ParentData', {})
        self.setdefault('ParentFlag', False)
        # 0 jobs are valid where we need to accept all blocks (only dqm etc subscriptions will run)
        self.setdefault('Jobs', 0)
        self.setdefault('WMSpec', None)
        self.setdefault('SiteWhitelist', [])
        self.setdefault('SiteBlacklist', [])
        self.setdefault('Dbs', None)
        self.setdefault('Task', None)
        self.setdefault('ParentQueueId', None)
        self.setdefault('Priority', 0)
        self.setdefault('SubscriptionId', None)
        self.setdefault('Status', None)
        self.setdefault('EventsWritten', 0)
        self.setdefault('FilesProcessed', 0)
        self.setdefault('PercentComplete', 0)
        self.setdefault('PercentSuccess', 0)
        self.setdefault('RequestName', None)
        self.setdefault('TaskName', None)
        self.setdefault('TeamName', None)
        self.setdefault('StartPolicy', {})
        self.setdefault('EndPolicy', {})
        self.setdefault('ACDC', {})
        self.setdefault('ChildQueueUrl', None)
        self.setdefault('ParentQueueUrl', None)
        self.setdefault('WMBSUrl', None)
        self.setdefault('NumberOfLumis', 0)
        self.setdefault('NumberOfEvents', 0)
        self.setdefault('NumberOfFiles', 0)
        # Number of files added to WMBS including parent files for this element. used only for monitoring purpose
        self.setdefault('NumOfFilesAdded', 0)
        # Mask used to constrain MC run/lumi ranges
        self.setdefault('Mask', None)
        # is new data being added to the workflow (e.g., growing input datasets)
        self.setdefault('OpenForNewData', False)
        # EPOCH time for when the last block was found and added to the workflow
        self.setdefault('TimestampFoundNewData', int(time.time()))
        # Trust initial input dataset location only or not
        self.setdefault('NoInputUpdate', False)
        # Trust initial pileup dataset location only or not
        self.setdefault('NoPileupUpdate', False)
        # EPOCH time with the creation time for this wq element, needed for sorting
        self.setdefault('CreationTime', int(time.time()))
        # set to true when updated from a WorkQueueElementResult
        self.modified = False

    def __to_json__(self, thunker):
        """Strip unthunkable"""
        # result = WorkQueueElement(thunker_encoded_json = True,
        result = dict(thunker_encoded_json=True,
                      type='WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement')
        result['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'] = {}
        result['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'].update(self)
        # Do we need this 'Subscription' or can we not store this at all?
        result['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'].pop('Subscription', None)
        result['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'].pop('WMSpec', None)
        #        if self.get('Id'):
        #            result['_id'] = result['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'].pop('Id')
        #        if self.get('_rev'):
        #            result['_rev'] = result['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'].pop('_rev')
        # result['Mask'] = thunker._thunk(result['Mask'])
        return result

    @property
    def id(self):
        """Generate id for element

        id is deterministic and can be used to identify duplicate elements.
        Calculation only includes fields which affect the workflow and input data.
        Result is an md5 hash of a ';' separated list of:
        workflow name, task name, list of inputs, mask, ACDC info, Dbs instance.

        Parent file info not accounted.

        Example:
        >>> WorkQueueElement(RequestName = 'a', TaskName = 'b').id
        '9ef03a6ad8f16d74fb5ba44df92bf1ef'

        Warning: Any change to this function may prevent identical existing and
        new elements from appearing equivalent, thus in the case of expanding
        work subscriptions work duplication can occur. Care must be taken
        if any modification is made.
        """
        if self._id:
            return self._id
        # Assume md5 is good enough for now
        myhash = md5()
        spacer = ';'  # character not present in any field
        myhash.update(encodeUnicodeToBytesConditional(self['RequestName'] + spacer, condition=PY3))
        # Task will be None in global inbox
        myhash.update(encodeUnicodeToBytesConditional(repr(self['TaskName']) + spacer, condition=PY3))
        myhash.update(encodeUnicodeToBytesConditional(",".join(sorted(self['Inputs'].keys())) + spacer, condition=PY3))
        # Check repr is reproducible - should be
        if self['Mask']:
            myhash.update(encodeUnicodeToBytesConditional(",".join(["%s=%s" % (x, y) for x, y in viewitems(self['Mask'])]) + spacer, condition=PY3))
        else:
            myhash.update(encodeUnicodeToBytesConditional("None" + spacer, condition=PY3))
        # Check ACDC is deterministic and all params relevant
        myhash.update(encodeUnicodeToBytesConditional(",".join(["%s=%s" % (x, y) for x, y in viewitems(self['ACDC'])]) + spacer, condition=PY3))
        myhash.update(encodeUnicodeToBytesConditional(repr(self['Dbs']) + spacer, condition=PY3))
        self._id = myhash.hexdigest()
        return self._id

    @id.setter
    def id(self, value):
        """Set id - use to override built-in id calculation"""
        self._id = value

    def __from_json__(self, jsondata, thunker):
        """"""
        self.update(jsondata)
        return self

    #        self.update(jsondata['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'])
    #        self.pop('type', None)
    #        self.pop('thunker_encoded_json', None)
    ##        self['Id'] = jsondata['_id']
    #        self['_rev'] = jsondata['_rev'] # what to do here???
    #        return self

    def inEndState(self):
        """Have we finished processing"""
        return self.isComplete() or self.isFailed() or self.isCanceled()

    def isComplete(self):
        return self['Status'] == 'Done'

    def isFailed(self):
        return self['Status'] == 'Failed'

    def isRunning(self):
        return self['Status'] == 'Running'

    def isAcquired(self):
        return self['Status'] == 'Acquired'

    def isAvailable(self):
        return self['Status'] == 'Available'

    def isCanceled(self):
        return self['Status'] == 'Canceled'

    def isCancelRequested(self):
        return self['Status'] == 'CancelRequested'

    def updateFromSubscription(self, wmbsStatus):
        """Get subscription status"""
        mapping = {'EventsWritten': 'events_written',
                   'FilesProcessed': 'files_processed',
                   'PercentComplete': 'percent_complete',
                   'PercentSuccess': 'percent_success'}
        for ourkey, wmbskey in viewitems(mapping):
            if wmbskey in wmbsStatus and self[ourkey] != wmbsStatus[wmbskey]:
                self['Modified'] = True
                self[ourkey] = wmbsStatus[wmbskey]

    def updateWithResult(self, progressReport):
        """Take a progress report and update ourself
           Return True if progress updated"""
        progressValues = ('Status', 'EventsWritten', 'FilesProcessed',
                          'PercentComplete', 'PercentSuccess')
        modified = False
        for val in progressValues:
            # ignore if values the same
            if self[val] == progressReport[val]:
                continue
            # ignore new state if it is lower than the current state
            if val == 'Status' and self[val] and STATES.index(progressReport[val]) <= STATES.index(self[val]):
                continue

            self[val] = progressReport[val]
            modified = True
        self.modified = modified

    def statusMetrics(self):
        """Returns the status & performance metrics"""
        return dict(Status=self['Status'],
                    PercentComplete=self['PercentComplete'],
                    PercentSuccess=self['PercentSuccess'])

    def passesSiteRestriction(self, site):
        """
        _passesSiteRestriction_

        Given a site name, checks whether it passes the site and data restrictions.
        """
        # Site list restrictions
        if site in self['SiteBlacklist'] or site not in self['SiteWhitelist']:
            return False

        # input data restrictions (TrustSitelists flag)
        if self['NoInputUpdate'] is False:
            for locations in viewvalues(self['Inputs']):
                if site not in locations:
                    return False
            if self['ParentFlag']:
                for locations in viewvalues(self['ParentData']):
                    if site not in locations:
                        return False

        # pileup data restrictions (TrustPUSitelists flag)
        if self['NoPileupUpdate'] is False:
            for locations in viewvalues(self['PileupData']):
                if site not in locations:
                    return False

        return True



#!/usr/bin/env python
"""
_GetUnsubscribedDatasets_

MySQL implementation of PhEDExInjector.Database.GetUnsubscribedDatasets
"""
from __future__ import division

from WMCore.Database.DBFormatter import DBFormatter


class GetUnsubscribedDatasets(DBFormatter):
    """
    _GetUnsubscribedDatasets_

    Gets the unsubscribed datasets from DBSBuffer
    """

    sql = """SELECT DISTINCT dbsbuffer_dataset_subscription.id,
                             dbsbuffer_dataset.path,
                             dbsbuffer_dataset_subscription.site,
                             dbsbuffer_dataset_subscription.custodial,
                             dbsbuffer_dataset_subscription.priority,
                             dbsbuffer_dataset_subscription.dataset_lifetime
               FROM dbsbuffer_dataset_subscription
               INNER JOIN dbsbuffer_dataset ON
                   dbsbuffer_dataset.id = dbsbuffer_dataset_subscription.dataset_id
               INNER JOIN dbsbuffer_algo_dataset_assoc ON
                 dbsbuffer_dataset.id = dbsbuffer_algo_dataset_assoc.dataset_id
               INNER JOIN dbsbuffer_file ON
                 dbsbuffer_algo_dataset_assoc.id = dbsbuffer_file.dataset_algo
             WHERE dbsbuffer_dataset_subscription.subscribed = 0 AND
                   dbsbuffer_file.in_phedex = 1 AND
                   dbsbuffer_dataset.path != 'bogus'"""

    def formatToPhEDEx(self, result):
        """
        _formatToPhEDEx_

        Format the result to the same format
        as the PhEDEx datasvc uses
        """
        for entry in result:
            entry['request_only'] = 'y'
            if entry['custodial'] == 1:
                entry['custodial'] = 'y'
            else:
                entry['custodial'] = 'n'
            entry['priority'] = entry['priority'].lower()
        return result

    def execute(self, conn=None, transaction=False):

        result = self.dbi.processData(self.sql, conn=conn,
                                      transaction=transaction)
        return self.formatToPhEDEx(self.formatDict(result))



#!/usr/bin/env python
"""
_DBSReader_

Readonly DBS Interface

"""

from WMCore.Services.DBS.DBSErrors import DBSReaderError
from WMCore.Services.DBS.DBS3Reader import DBS3Reader


def DBSReader(endpoint, **kwargs):
    """Function to find and instantiate desired DBSReader object"""

    try:
        dbs = DBS3Reader(endpoint, **kwargs)
        # if this doesn't throw endpoint is dbs3
        dbs.dbs.serverinfo()
        return dbs
    except Exception as ex:
        msg = 'Instantiating DBS3Reader failed with %s\n' % str(ex)
        raise DBSReaderError("Can't contact DBS at %s, got errors %s" % (endpoint, msg))



#!/usr/bin/env python
"""
_WMComponent_

Contains generic components that can be
used cross project.

Project specific component should be developed
within its own project space (not in wmcomponent).

"""
__all__ = []



#!/usr/bin/env python
"""
_LoadForMonitoring_

Oracle implementation for loading a job by scheduler status
"""


from WMCore.BossAir.MySQL.LoadForMonitoring import LoadForMonitoring as MySQLLoadForMonitoring

class LoadForMonitoring(MySQLLoadForMonitoring):
    """
    _LoadForMonitoring_

    Load all jobs with a certain scheduler status including
    all the joined information.
    """



#!/usr/bin/env python
"""
_GetSite_

Oracle implementation of JobGroup.GetSite
"""

__all__ = []



from WMCore.WMBS.MySQL.JobGroup.GetSite import GetSite as MySQLGetSite

class GetSite(MySQLGetSite):
    """
    Right now, same as MySQL

    """



#!/usr/bin/env python
"""
_WorkQueueManagerPoller_

Pull work out of the work queue.
"""
import logging
import random
import threading
import time

from Utils.Timers import timeFunction
from WMComponent.JobSubmitter.JobSubmitAPI import availableScheddSlots
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.PyCondor.PyCondorAPI import PyCondorAPI
from WMCore.Services.ReqMgrAux.ReqMgrAux import isDrainMode
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread


class WorkQueueManagerWorkPoller(BaseWorkerThread):
    """
    Polls for Work
    """

    def __init__(self, queue, config):
        """
        Initialise class members
        """
        BaseWorkerThread.__init__(self)
        myThread = threading.currentThread()

        self.queue = queue
        self.config = config
        self.condorAPI = PyCondorAPI()

        self.daoFactory = DAOFactory(package="WMCore.WMBS", logger=logging, dbinterface=myThread.dbi)
        self.listSubsWithoutJobs = self.daoFactory(classname="Subscriptions.GetSubsWithoutJobGroup")

    def setup(self, parameters):
        """
        Called at startup - introduce random delay
             to avoid workers all starting at once
        """
        t = random.randrange(self.idleTime)
        self.logger.info('Sleeping for %d seconds before 1st loop', t)
        time.sleep(t)

    @timeFunction
    def algorithm(self, parameters):
        """
        Pull in work
        """
        self.logger.info("Starting WorkQueueManagerWorkPoller thread ...")
        try:
            self.pullWork()
        except Exception as ex:
            self.queue.logger.error("Error in work pull loop: %s", str(ex))
        try:
            # process if we get work or not - we may have to split old work
            # i.e. if transient errors were seen during splitting
            self.processWork()
        except Exception as ex:
            self.queue.logger.error("Error in new work split loop: %s", str(ex))
        return

    def passRetrieveCondition(self):
        """
        _passRetrieveCondition_
        Return true if the component can proceed with fetching work.
        False if the component should skip pulling work this cycle.

        For now, it only checks whether the agent is in drain mode or
        MAX_JOBS_PER_OWNER is reached or if the condor schedd is overloaded.
        """
        passCond = "OK"
        myThread = threading.currentThread()
        if isDrainMode(self.config):
            passCond = "agent is in drain mode"
        elif availableScheddSlots(myThread.dbi) <= 0:
            passCond = "schedd slot is maxed: MAX_JOBS_PER_OWNER"
        elif self.condorAPI.isScheddOverloaded():
            passCond = "schedd is overloaded"
        else:
            subscriptions = self.listSubsWithoutJobs.execute()
            if subscriptions:
                passCond = "JobCreator hasn't created jobs for subscriptions %s" % subscriptions

        return passCond

    def pullWork(self):
        """Get work from parent"""
        self.queue.logger.info("Pulling work from %s", self.queue.parent_queue.queueUrl)

        myThread = threading.currentThread()

        try:
            cond = self.passRetrieveCondition()
            if cond == "OK":
                work = self.queue.pullWork()
                self.queue.logger.info("Obtained %s unit(s) of work", work)
                myThread.logdbClient.delete("LocalWorkQueue_pullWork", "warning", this_thread=True)
            else:
                self.queue.logger.warning("No work will be pulled, reason: %s", cond)
                myThread.logdbClient.post("LocalWorkQueue_pullWork", cond, "warning")
        except IOError as ex:
            self.queue.logger.exception("Error opening connection to work queue: %s", str(ex))
        except Exception as ex:
            self.queue.logger.exception("Unable to pull work from parent Error: %s", str(ex))

    def processWork(self):
        """Process new work"""
        self.queue.logger.info("Splitting new work")
        try:
            self.queue.processInboundWork()
        except Exception as ex:
            self.queue.logger.exception('Error during split: %s', str(ex))
        self.logger.info('Splitting finished')
        return



#!/usr/bin/env python
"""
_SucceededJobs_

Oracle implementation of Subscriptions.SucceededJobs
"""




from WMCore.WMBS.MySQL.Subscriptions.SucceededJobs import SucceededJobs as SucceededJobsMySQL

class SucceededJobs(SucceededJobsMySQL):
    pass



#!/usr/bin/env python
"""
_NewDataset_

Oracle implementation of DBS3Buffer.NewDataset
"""




from WMComponent.DBS3Buffer.MySQL.NewDataset import NewDataset as MySQLNewDataset

class NewDataset(MySQLNewDataset):

    sql = """INSERT INTO dbsbuffer_dataset
             (id, path, processing_ver, acquisition_era, valid_status, global_tag, parent, prep_id)
             SELECT dbsbuffer_dataset_seq.nextval, :path, :processing_ver, :acquisition_era,
                    :valid_status, :global_tag, :parent, :prep_id
             FROM DUAL
             WHERE NOT EXISTS
               ( SELECT *
                 FROM dbsbuffer_dataset
                 WHERE path = :path )
             """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module which constructs the ReqMgr2 authorization setup,
defining which CRIC roles and groups can perform which
actions, based on the target request status.

It defines 3 different permissions, such as:
* admin: can perform any status transition and non-status write action
* ops: can perform many status transition and non-status write action,
    but not all of them (service-based actions are restricted).
* ppd: can perform only a few status transition and non-status write
   action. Service and Ops-based actions are forbidden

The 2 data structures required for this module are defined in the
ReqMgr2 service configuration file, a short explanation of them follows:
 * authzByStatus: a list with 3 dictionaries in it. Each dictionary
     defines a "permission" level and which target statuses are allowed.
     E.g.:
         [{"permission": "admin",
           "statuses": ["acquired", "running-open"]},
          {"permission": "ops",
           "statuses": ["assigned", "staging", "staged"]},
          {"permission": "ppd",
           "statuses": ["new", "aborted", "NO_STATUS"]}]
 * authzRolesGroups: a dictionary defining which CRIC roles/groups
     are member of a given permission level. E.g.:
         {"admin": {'role': ["role_1"], 'group': ["group_1"]},
          "ops": {'role': ["role_1", "role_2"], 'group': ["group_1", "group_2"]},
          "ppd": {'role': ["role_1", "role_2", "role_3"], 'group': ["group_1", "group_2", "group_3"]}}

NOTE that "NO_STATUS" defines who is allowed to perform a write action
not involving request status transition
"""
from copy import deepcopy

from WMCore.ReqMgr.DataStructs.RequestStatus import ALLOWED_ACTIONS_FOR_STATUS


class AuthzByStatus:
    """
    Class defining the authorization policy based on the
    target request status.

    This class depends on the ReqMgr2 authorization configuration.
    """

    def __init__(self, authzByStatus, authzRolesGroups):
        """
        Create a request status based data structure, with the role and
        groups passed from the service configuration file.
        :param authzByStatus: a list of dictionary items. Where each dict contains
            a permission group and a list of allowed statuses.
        :param authzRolesGroups: a nested dictionary with CRIC roles and groups
            permissions for each permissions group
        :return: None
        """
        self.noStatus = None
        self.authzByStatus = {}
        self.authzRolesGroups = authzRolesGroups

        # first, map status to required permission
        for item in authzByStatus:
            for status in item["statuses"]:
                if status == "NO_STATUS":
                    self.noStatus = item["permission"]
                else:
                    self.authzByStatus[status] = item["permission"]

        # now validate the authorization configuration
        self._validateAuthzStruct()

    def _validateAuthzStruct(self):
        """
        Runs a basic validation of the authorization data structure
        to catch possible mistakes made in the configuration file.
        :return: raises an exception if problems are found
        """
        if self.noStatus is None:
            msg = "Configuration for authorization by status does not define the required "
            msg += "roles and groups for a non-status change update. Please review it."
            raise RuntimeError(msg)

        expectedPermissions = set(["admin", "ops", "ppd"])
        if set(list(self.authzRolesGroups)) != expectedPermissions:
            msg = "Authorized roles and groups do not match the expected group permissions, "
            msg += f"they are: {expectedPermissions}. Please review it."
            raise RuntimeError(msg)

        if set(self.authzByStatus.values()) != expectedPermissions:
            msg = "Configuration for authorization by status does not match the expected "
            msg += f"group permissions, they are: {expectedPermissions}. Please review it."
            raise RuntimeError(msg)

        if len(list(ALLOWED_ACTIONS_FOR_STATUS.keys())) != len(self.authzByStatus.keys()):
            msg = "Configuration for authorization by status is missing workflow "
            msg += "status(es). Please review it and make sure all statuses are covered."
            raise RuntimeError(msg)

    def getRolesGroupsByStatus(self, requestArgs):
        """
        Given a dictionary with the request arguments being changed by
        the user, decide which roles/groups the user needs to have in
        order to complete the desired action.
        :param requestArgs: dictionary with request arguments posted by the user
        :return: dictionary with the roles and groups required for such action
        """
        targetStatus = requestArgs.get("RequestStatus", None)
        if targetStatus is None:
            return deepcopy(self.authzRolesGroups[self.noStatus])

        permissionGroup = self.authzByStatus[targetStatus]
        return deepcopy(self.authzRolesGroups[permissionGroup])



#!/usr/bin/env python
"""
_GetState_

Oracle implementation of Jobs.GetState
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.GetState import GetState as MySQLGetState

class GetState(MySQLGetState):
    """
    Right now this is the same as the MySQL version.

    """



#!/usr/bin/env python
"""
_CountBlocks_

MySQL implementation of DBSBuffer.CountBlocks
"""




from WMCore.Database.DBFormatter import DBFormatter

class CountBlocks(DBFormatter):
    """
    _CountBlocks_

    Count the number of blocks in the buffer.  This is used by the DBSUpload
    unit test.
    """
    sql = "SELECT COUNT(*) FROM dbsbuffer_block"

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, conn = conn,
                                          transaction = transaction)
        result = self.format(result)
        return result[0][0]



#!/usr/bin/env python
"""
_SplitFileBased_

JobSplitting algorithm for creating jobs that run over the results of split
processing jobs.  This borrows heavily from the generic WMBS merging algorithm
but instead of combining multiple merge units together this will create a
single job for each merge unit.
"""

import threading
from functools import cmp_to_key

from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File


def mergeUnitCompare(a, b):
    """
    _mergeUnitCompare_

    Compare two merge units.  They will be sorted first by run ID and then by
    lumi ID.
    """
    if a["run"] > b["run"]:
        return 1
    elif a["run"] == b["run"]:
        if a["lumi"] > b["lumi"]:
            return 1
        elif a["lumi"] == b["lumi"]:
            return 0
        else:
            return -1
    else:
        return -1


def fileCompare(a, b):
    """
    _fileCompare_

    Compare two files based on their "file_first_event" attribute.
    """
    if a["file_first_event"] > b["file_first_event"]:
        return 1
    if a["file_first_event"] == b["file_first_event"]:
        return 0
    else:
        return -1


def sortedFilesFromMergeUnits(mergeUnits):
    """
    _sortedFilesFromMergeUnits_

    Given a list of merge units sort them and the files that they contain.
    Return a list of sorted WMBS File structures.
    """
    mergeUnits.sort(key=cmp_to_key(mergeUnitCompare))
    sortedFiles = []
    for mergeUnit in mergeUnits:
        mergeUnit["files"].sort(key=cmp_to_key(fileCompare))
        for file in mergeUnit["files"]:
            newFile = File(id=file["file_id"], lfn=file["file_lfn"],
                           events=file["file_events"])
            newFile.addRun(Run(file["file_run"], file["file_lumi"]))

            # The WMBS data structure puts locations that are passed in through
            # the constructor in the "newlocations" attribute.  We want these to
            # be in the "locations" attribute so that they get picked up by the
            # job submitter.
            newFile["locations"] = set([file["pnn"]])
            newFile.addRun(Run(file["file_run"], file["file_lumi"]))
            sortedFiles.append(newFile)

    return sortedFiles


class SplitFileBased(JobFactory):
    """
    _SplitFileBased_

    JobSplitting algorithm for creating jobs that run over the results of split
    processing jobs.
    """

    def defineMergeUnits(self, mergeableFiles):
        """
        _defineMergeUnits_

        Split all the mergeable files into merge units.  A merge unit is a group
        of files that must be merged together.  For example, the files that
        result from event based splitting jobs need to be merged back together.
        This method will return a list of merge units.  A merge unit is a
        dictionary with the following keys: file_parent, total_events, total_size,
        run, lumi, and files.  The files in the merge group are stored in a list
        under the files key.
        """
        mergeUnits = {}
        newMergeUnit = {}

        for mergeableFile in mergeableFiles:
            newMergeFile = {}
            for key in mergeableFile:
                newMergeFile[key] = mergeableFile[key]

            if newMergeFile["file_run"] not in mergeUnits:
                mergeUnits[newMergeFile["file_run"]] = []

            for mergeUnit in mergeUnits[newMergeFile["file_run"]]:
                if mergeUnit["file_parent"] == mergeableFile["file_parent"]:
                    mergeUnit["files"].append(newMergeFile)
                    mergeUnit["total_size"] += newMergeFile["file_size"]
                    mergeUnit["total_events"] += newMergeFile["file_events"]

                    if mergeableFile["file_run"] < mergeUnit["run"] or \
                            (mergeableFile["file_run"] == mergeUnit["run"] and \
                                         mergeableFile["file_lumi"] < mergeUnit["lumi"]):
                        newMergeUnit["run"] = newMergeFile["file_run"]
                        newMergeUnit["lumi"] = newMergeFile["file_lumi"]

                    break
            else:
                newMergeUnit = {}
                newMergeUnit["file_parent"] = newMergeFile["file_parent"]
                newMergeUnit["total_events"] = newMergeFile["file_events"]
                newMergeUnit["total_size"] = newMergeFile["file_size"]
                newMergeUnit["run"] = newMergeFile["file_run"]
                newMergeUnit["lumi"] = newMergeFile["file_lumi"]
                newMergeUnit["files"] = []
                newMergeUnit["files"].append(newMergeFile)
                mergeUnits[newMergeFile["file_run"]].append(newMergeUnit)

        return mergeUnits

    def createProcJobs(self, mergeUnits):
        """
        _createProcJob_

        Given a list of merge units, create a job that processes all the files
        in the merge unit.  Each job will be placed in its own JobGroup so that
        merging will work correctly.
        """
        for mergeUnit in mergeUnits:
            self.newGroup()
            self.newJob(name=makeUUID())
            sortedFiles = sortedFilesFromMergeUnits([mergeUnit])

            for file in sortedFiles:
                file.load()
                self.currentJob.addFile(file)

    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        Use the generic WMBS merging DAO to get a list of files in our fileset
        that correspond to completed job groups.  Create jobs for these files.
        """
        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)

        mergeDAO = daoFactory(classname="Subscriptions.GetFilesForMerge")
        mergeableFiles = mergeDAO.execute(self.subscription["id"])

        mergeUnits = self.defineMergeUnits(mergeableFiles)
        for runNumber in mergeUnits:
            mergeUnits[runNumber].sort(key=cmp_to_key(mergeUnitCompare))
            self.createProcJobs(mergeUnits[runNumber])

        return self.jobGroups



#!/usr/bin/env python
"""
_CountFilesByStatus_

Oracle implementation of DBSBuffer.CountFilesByStatus
"""

from __future__ import division
from WMComponent.DBS3Buffer.MySQL.CountFilesByStatus import CountFilesByStatus as MySQLCountFilesByStatus

class CountFilesByStatus(MySQLCountFilesByStatus):
    """
    CountFilesByStatus

    """
    pass



#!/usr/bin/env python
"""
WorkQueue End Policy

"""

from WMCore.WMFactory import WMFactory

endFac = WMFactory(__name__, __name__)


def endPolicy(elements, parents=None, args=None):
    """Load an end policy"""

    # load policy and apply
    if not args:
        args = {}
    if elements:
        policy = elements[0]['EndPolicy']
    elif parents:
        policy = parents[0]['EndPolicy']
    else:
        raise RuntimeError("Can't get policy, no elements or parents")
    name = policy['policyName']
    args = args.get(name, {})
    args.update(policy)
    return endFac.loadObject(name,
                             args,
                             storeInCache=False)(elements, parents)


__all__ = []



#!/usr/bin/env python
"""
_SiblingSubscriptionsFailed_

Oracle implementation of Subscription.SiblingSubscriptionsFailed
"""

from WMCore.WMBS.MySQL.Subscriptions.SiblingSubscriptionsFailed import \
    SiblingSubscriptionsFailed as SiblingFailedMySQL

class SiblingSubscriptionsFailed(SiblingFailedMySQL):

    insert = """INSERT INTO wmbs_sub_files_complete (subscription, fileid)
                  SELECT :subscription, :fileid FROM dual
                  WHERE NOT EXISTS (SELECT wsfc.fileid FROM wmbs_sub_files_complete wsfc
                                    WHERE subscription = :subscription
                                    AND fileid = :fileid)"""



#!/usr/bin/env python
"""
_couch-trash_

"""

from builtins import range
import os
import random
import time
import sys

from WMCore.Configuration import loadConfigurationFile
from WMCore.WMInit import connectToDB
from WMCore.WMInit import getWMBASE

from WMCore.Services.UUIDLib import makeUUID
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.DataStructs.Job import Job
from WMCore.DataStructs.File import File
from WMCore.FwkJobReport.Report import Report

lastJobID = int(sys.argv[1])

def createFile():
    """
    _createFile_

    Create a file with some random metdata.
    """
    newFile = File(lfn = makeUUID(), size = random.randrange(1024, 1048576, 1024),
                   events = random.randrange(10, 100000, 50),
                   parents = [File(lfn = makeUUID())],
                   locations = makeUUID())
    newFile["first_event"] = 0
    newFile["last_event"] = 0
    newFile["id"] = 1
    return newFile

def createJobs(totalJobs = 100):
    """
    _createJobs_

    """
    global lastJobID

    newJobs = []
    for i in range(totalJobs):
        newJob = Job(name = makeUUID(), files = [createFile(), createFile()])
        newJob["task"] = "SomeTask"
        newJob["workflow"] = "SomeWorkflow"
        newJob["owner"] = "sfoulkes@fnal.gov"
        newJob["id"] = lastJobID
        newJob["couch_record"] = None
        newJobs.append(newJob)
        lastJobID += 1

    return newJobs

def thrashCouch():
    """
    _thrashCouch_

    """
    jobs = {"new": set(), "created": set(), "executing": set(),
            "complete": set(), "success": set(), "cleanout": set()}

    config = loadConfigurationFile(os.environ["WMAGENT_CONFIG"])
    changeState = ChangeState(config)

    myReport = Report()
    myReport.unpersist(os.path.join(getWMBASE(), "test/python/WMComponent_t/JobAccountant_t/fwjrs/LoadTest00.pkl"))

    for i in range(500):
        jobs = createJobs()
        changeState.recordInCouch(jobs, "created", "new")
        changeState.recordInCouch(jobs, "executing", "created")
        changeState.recordInCouch(jobs, "complete", "executing")

        for job in jobs:
            job["fwjr"] = myReport

        changeState.recordInCouch(jobs, "success", "complete")

        for job in jobs:
            job["fwjr"] = None

        changeState.recordInCouch(jobs, "cleanout", "success")
        #time.sleep(10)
    return

connectToDB()
thrashCouch()



#!/usr/bin/env python
"""
_Executor_

Interface definition for a step executor


"""
from __future__ import absolute_import
from builtins import object

import json
import logging
import os
import subprocess
import sys

from Utils.FileTools import getFullPath
from Utils.PythonVersion import PY3
from Utils.Utilities import zipEncodeStr, decodeBytesToUnicodeConditional
from WMCore.FwkJobReport.Report import Report
from WMCore.WMSpec.Steps.StepFactory import getStepEmulator
from WMCore.WMSpec.WMStep import WMStepHelper

getStepName = lambda step: WMStepHelper(step).name()
getStepErrorDestination = lambda step: WMStepHelper(step).getErrorDestinationStep()


def getStepSpace(stepName):
    """
    _getStepSpace_

    Util to get the runtime step space.
    This imports dynamic runtime libraries so be careful how
    you use it

    """
    modName = "WMTaskSpace"
    if modName in sys.modules:
        taskspace = sys.modules[modName]
    else:
        try:
            # taskspace = __import__(modName, globals(), locals(), ['taskSpace'], -1)
            taskspace = __import__(modName, globals(), locals(), ['taskSpace'])

        except ImportError as ex:
            msg = "Unable to load WMTaskSpace module:\n"
            msg += str(ex)
            # TODO: Generic ExecutionException...
            raise RuntimeError(msg)

    try:
        stepSpace = taskspace.taskSpace.stepSpace(stepName)
    except Exception as ex:
        msg = "Error retrieving stepSpace from TaskSpace:\n"
        msg += str(ex)
        raise RuntimeError(msg)
    return stepSpace


class Executor(object):
    """
    _Executor_

    Define API for a step during execution

    """

    def __init__(self):
        self.report = None
        self.diagnostic = None
        self.emulator = None
        self.emulationMode = False
        self.step = None
        self.stepName = None
        self.stepSpace = None
        self.task = None
        self.workload = None
        self.job = None
        self.errorDestination = None
        self.logger = logging.getLogger()
        self.logger.info("Steps.Executor logging started")

    def initialise(self, step, job):
        """
        _initialise_


        Initialise the executor attributes

        """
        self.step = step
        self.job = job
        self.stepName = getStepName(self.step)
        self.stepSpace = getStepSpace(self.stepName)
        self.task = self.stepSpace.getWMTask()
        self.workload = self.stepSpace.taskSpace.workload
        self.report = Report(self.stepName)
        self.report.data.task = self.task.name()
        self.report.data.workload = self.stepSpace.taskSpace.workloadName()
        self.report.data.id = job['id']
        self.errorDestination = getStepErrorDestination(self.step)

        self.step.section_("execution")
        self.step.execution.exitStatus = 0
        self.step.execution.reportLocation = "%s/Report.pkl" % (
            self.stepSpace.location,
        )

        # Set overall step status to 1 (failed)
        self.report.setStepStatus(stepName=self.stepName, status=1)

        #  //
        # //  Does the step contain settings for an emulator?
        # //   If so, load it up

        emulatorName = getattr(self.step.emulator, "emulatorName", None)
        if emulatorName != None:
            self.emulator = getStepEmulator(emulatorName)
            self.emulator.initialise(self)
            self.emulationMode = True

        return

    def saveReport(self):
        """
        _saveReport_

        Save the job report

        """
        self.report.persist(self.step.execution.reportLocation)
        return

    def pre(self, emulator=None):
        """
        _pre_

        pre execution checks. Can alter flow of execution by returning
        a different step in the task. If None, then current step will
        be passed to execute.

        TODO: Define better how to switch to different step within the task

        """
        return None

    def execute(self, emulator=None):
        """
        _execute_

        Override behaviour to execute this step type.
        If Emulator is provided, execute the emulator instead.
        Return a framework job report instance

        """
        msg = "WMSpec.Steps.Executor.execute method not overridden in "
        msg += "implementation: %s\n" % self.__class__.__name__
        raise NotImplementedError(msg)

    def post(self, emulator=None):
        """
        _post_

        post execution checks. Can alter flow of execution by returning
        a different step in the task. If None, then the next step in the task
        will be used next.

        TODO: Define better how to switch to different step within the task

        """
        return None

    def setCondorChirpAttrDelayed(self, key, value, compress=False, maxLen=5120):
        """
        _setCondorChirpAttrDelayed_

        Util to call condor_chirp and publish the key/value pair

        """

        if compress:
            value = zipEncodeStr(value, maxLen=maxLen)

        # bytes object is not JSON serializable in Python3
        value = decodeBytesToUnicodeConditional(value, condition=PY3)
        # construct condor_chirp binary location from CONDOR_CONFIG
        # Note: This works when we do not use containers.
        condor_chirp_bin = None
        condor_config = os.getenv('CONDOR_CONFIG', None)
        if condor_config:
            condor_config_dir = os.path.dirname(condor_config)
            condor_chirp_bin = os.path.join(condor_config_dir, 'main/condor/libexec/condor_chirp')

        # If the above fails, look for the executable in the environment
        # This is the usual case for containers
        if not condor_chirp_bin or not os.path.isfile(condor_chirp_bin):
            condor_chirp_bin = getFullPath("condor_chirp")

        if condor_chirp_bin and os.access(condor_chirp_bin, os.X_OK):
            args = [condor_chirp_bin, 'set_job_attr_delayed', key, json.dumps(value)]
            subprocess.call(args)
        else:
            if condor_chirp_bin and not os.access(condor_chirp_bin, os.X_OK):
                msg = 'condor_chirp was found in: %s, but it was not an executable.' % condor_chirp_bin
            else:
                msg = 'condor_chirp was not found in the system.'
            self.logger.warning(msg)

        return



#!/usr/bin/env python
"""
_DBSUtils_

set of common utilities for DBS3Reader

"""
import json
import urllib
from urllib.parse import urlparse, parse_qs, quote_plus
from collections import defaultdict

from Utils.CertTools import cert, ckey
from dbs.apis.dbsClient import aggFileLumis, aggFileParents
from WMCore.Services.pycurl_manager import getdata as multi_getdata
from Utils.PortForward import PortForward


def dbsListFileParents(dbsUrl, blocks):
    """
    Concurrent counter part of DBS listFileParents API

    :param dbsUrl: DBS URL
    :param blocks: list of blocks
    :return: list of file parents
    """
    urls = ['%s/fileparents?block_name=%s' % (dbsUrl, quote_plus(b)) for b in blocks]
    func = aggFileParents
    uKey = 'block_name'
    return getUrls(urls, func, uKey)


def dbsListFileLumis(dbsUrl, blocks):
    """
    Concurrent counter part of DBS listFileLumis API

    :param dbsUrl: DBS URL
    :param blocks: list of blocks
    :return: list of file lumis
    """
    urls = ['%s/filelumis?block_name=%s' % (dbsUrl, quote_plus(b)) for b in blocks]
    func = aggFileLumis
    uKey = 'block_name'
    return getUrls(urls, func, uKey)


def dbsBlockOrigin(dbsUrl, blocks):
    """
    Concurrent counter part of DBS files API

    :param dbsUrl: DBS URL
    :param blocks: list of blocks
    :return: list of block origins for a given parent lfns
    """
    urls = ['%s/blockorigin?block_name=%s' % (dbsUrl, quote_plus(b)) for b in blocks]
    func = None
    uKey = 'block_name'
    return getUrls(urls, func, uKey)


def dbsParentFilesGivenParentDataset(dbsUrl, parentDataset, fInfo):
    """
    Obtain parent files for given fileInfo object

    :param dbsUrl: DBS URL
    :param parentDataset: parent dataset name
    :param fInfo: file info object
    :return: list of parent files for given file info object
    """
    portForwarder = PortForward(8443)
    urls = []
    for fileInfo in fInfo:
        run = fileInfo['run_num']
        lumis = urllib.parse.quote_plus(str(fileInfo['lumi_section_num']))
        url = f'{dbsUrl}/files?dataset={parentDataset}&run_num={run}&lumi_list={lumis}'
        urls.append(portForwarder(url))
    func = None
    uKey = None
    rdict = getUrls(urls, func, uKey)
    parentFiles = defaultdict(set)
    for fileInfo in fInfo:
        run = fileInfo['run_num']
        lumis = urllib.parse.quote_plus(str(fileInfo['lumi_section_num']))
        url = f'{dbsUrl}/files?dataset={parentDataset}&run_num={run}&lumi_list={lumis}'
        url = portForwarder(url)
        if url in rdict:
            pFileList = rdict[url]
            pFiles = {x['logical_file_name'] for x in pFileList}
            parentFiles[fileInfo['logical_file_name']] = \
                parentFiles[fileInfo['logical_file_name']].union(pFiles)
    return parentFiles


def getUrls(urls, aggFunc, uKey=None):
    """
    Perform parallel DBS calls for given set of urls and apply given aggregation
    function to the results.

    :param urls: list of DBS urls to call
    :param aggFunc: aggregation function
    :param uKey: url parameter to use for final dictionary
    :return: dictionary of resuls where keys are urls and values are obtained results
    """
    data = multi_getdata(urls, ckey(), cert())

    rdict = {}
    for row in data:
        url = row['url']
        code = int(row.get('code', 200))
        error = row.get('error')
        if code != 200:
            msg = f"Fail to query {url}. Error: {code} {error}"
            raise RuntimeError(msg)
        if uKey:
            key = urlParams(url).get(uKey)
        else:
            key = url
        data = row.get('data', [])
        res = json.loads(data)
        if aggFunc:
            rdict[key] = aggFunc(res)
        else:
            rdict[key] = res
    return rdict


def urlParams(url):
    """
    Return dictionary of URL parameters

    :param url: URL link
    :return: dictionary of URL parameters
    """
    parsedUrl = urlparse(url)
    rdict = parse_qs(parsedUrl.query)
    for key, vals in rdict.items():
        if len(vals) == 1:
            rdict[key] = vals[0]
    return rdict



#!/usr/bin/env python
"""
_HeritageLFNParent_

MySQL implementation of DBSBufferFiles.HeritageLFNParent
"""




from WMCore.Database.DBFormatter import DBFormatter

class HeritageLFNParent(DBFormatter):
    sql = """INSERT INTO dbsbuffer_file_parent (child, parent)
               SELECT dfc.id, dfp.id FROM dbsbuffer_file dfc
               INNER JOIN dbsbuffer_file dfp
               WHERE dfc.lfn = :child
               AND dfp.lfn = :lfn
    """


    #sql = """INSERT INTO dbsbuffer_file_parent (child, parent)
    #           SELECT :child, dbsbuffer_file.id FROM dbsbuffer_file
    #             WHERE dbsbuffer_file.lfn = :lfn AND NOT EXISTS
    #               (SELECT child FROM dbsbuffer_file_parent
    #                WHERE child = :child AND
    #                      parent = (SELECT id FROM dbsbuffer_file WHERE lfn = :lfn))"""

    def execute(self, parentLFNs, childLFN, conn = None, transaction = False):
        binds = []
        for parentLFN in parentLFNs:
            binds.append({"lfn": parentLFN, "child": childLFN})

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        result = self.dbi.processData("SELECT * FROM dbsbuffer_file_parent", None,
                                      conn = conn, transaction = transaction)

        return



#!/usr/bin/env python
"""
_PersistencyHelper_

Util class to provide a common persistency layer for ConfigSection derived
objects, with options to save in different formats

Placeholder for ideas at present....

"""
from __future__ import print_function

from future.moves.urllib.parse import urlparse
from future.moves.urllib.request import urlopen, Request

from builtins import object

import pickle


class PersistencyHelper(object):
    """
    _PersistencyHelper_

    Save a WMSpec object to a file using pickle

    Future ideas:
    - pickle mode: read/write using pickle
    - python mode: write using pythonise, read using import
       Needs work to preserve tree information
    - gzip mode: also gzip/unzip content if set to True
    - json mode: read/write using json

    """

    def save(self, filename):
        """
        _save_

        Save data to a file
        Saved format is defined depending on the extension
        """
        with open(filename, 'wb') as handle:
            # TODO: use different encoding scheme for different extension
            # extension = filename.split(".")[-1].lower()
            # FIXME: once both central services and WMAgent are in Py3, we can remove protocol=0
            pickle.dump(self.data, handle, protocol=0)
        return

    def load(self, filename):
        """
        _load_

        Unpickle data from file

        """

        # TODO: currently support both loading from file path or url
        # if there are more things to filter may be separate the load function

        # urllib2 needs a scheme - assume local file if none given
        if not urlparse(filename)[0]:
            filename = 'file:' + filename
            handle = urlopen(Request(filename, headers={"Accept": "*/*"}))
            self.data = pickle.load(handle)
            handle.close()
        elif filename.startswith('file:'):
            handle = urlopen(Request(filename, headers={"Accept": "*/*"}))
            self.data = pickle.load(handle)
            handle.close()
        else:
            # use own request class so we get authentication if needed
            from WMCore.Services.Requests import Requests
            request = Requests(filename)
            data = request.makeRequest('', incoming_headers={"Accept": "*/*"}, decoder=False)
            self.data = pickle.loads(data[0])

        return

    def saveCouch(self, couchUrl, couchDBName, metadata=None):
        """ Save this spec in CouchDB.  Returns URL """
        from WMCore.Database.CMSCouch import CouchServer, CouchInternalServerError
        metadata = metadata or {}
        server = CouchServer(couchUrl)
        database = server.connectDatabase(couchDBName)
        name = self.name()
        uri = '/%s/%s' % (couchDBName, name)
        specuri = uri + '/spec'
        if not database.documentExists(name):
            self.setSpecUrl(couchUrl + specuri)
            doc = database.put(uri, data=metadata, contentType='application/json')
            # doc = database.commitOne(self.name(), metadata)
            rev = doc['rev']
        else:
            # doc = database.get(uri+'?revs=true')
            doc = database.document(name)
            rev = doc['_rev']

        # specuriwrev = specuri + '?rev=%s' % rev
        # FIXME: once both central services and WMAgent are in Py3, we can remove protocol=0
        workloadString = pickle.dumps(self.data, protocol=0)
        # result = database.put(specuriwrev, workloadString, contentType='application/text')
        retval = database.addAttachment(name, rev, workloadString, 'spec')
        if retval.get('ok', False) is not True:
            msg = "Failed to save a spec attachment in CouchDB for %s" % name
            raise CouchInternalServerError(msg, data=None, result=retval)

        url = couchUrl + specuri
        return url

    def splitCouchUrl(self, url):
        """ Splits a URL into baseURL, dbname, and document """
        toks = url.split('/')
        dbname = toks[3]
        # assume that the name "couchdb" is a redirected URL
        if dbname == "couchdb":
            dbname = toks[4]
        toks = url.split('/%s/' % dbname)
        return toks[0], dbname, toks[1]

    def saveCouchUrl(self, url):
        """ Saves the spec to a given Couch URL """
        couchUrl, dbname, doc = self.splitCouchUrl(url)
        return self.saveCouch(couchUrl, dbname)

    def deleteCouch(self, couchUrl, couchDBName, id):
        import WMCore.Database.CMSCouch
        server = WMCore.Database.CMSCouch.CouchServer(couchUrl)
        database = server.connectDatabase(couchDBName)
        # doesn't work
        if not database.documentExists(id):
            print("Could not find document " + id)
            return
        doc = database.delete_doc(id)
        return



#!/usr/bin/env python
"""
_ListAlgo_

Oracle implementation of DBSBuffer.ListAlgo
"""




from WMComponent.DBS3Buffer.MySQL.ListAlgo import ListAlgo as MySQLListAlgo

class ListAlgo(MySQLListAlgo):
    """
    _ListAlgo_

    Retrieve information about an algorithm from the DBSBuffer.  This is mainly
    used by the unit tests to verify that the NewAlgo DAO is working correctly.
    """
    pass



#!/usr/bin/env python
"""
_GetFWJRByState_

MySQL implementation of Jobs.GetFWJRByState
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetFWJRByState(DBFormatter):
    """
    _GetFWJRByState_

    Retrieve the ID and framework job report path of all jobs in a particular
    state.
    """
    sql = """SELECT id, fwjr_path FROM wmbs_job WHERE state =
               (SELECT id FROM wmbs_job_state WHERE name = :state)"""

    def format(self, results):
        """
        _format_

        """
        results = DBFormatter.format(self, results)

        jobs = []
        for result in results:
            jobs.append({"id": result[0], "fwjr_path": result[1]})

        return jobs

    def execute(self, state, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"state": state}, conn = conn,
                                      transaction = transaction)

        return self.format(result)



#!/usr/bin/env python
"""
_GetSiteInfo_

MySQL implementation of Locations.GetSiteInfo
"""

from WMCore.Database.DBFormatter import DBFormatter


class GetSiteInfo(DBFormatter):
    """
    Grab all the relevant information for a given site.
    Usually useful only in the submitter
    """
    sql = """SELECT wl.site_name, wpnn.pnn, wl.ce_name, wl.pending_slots,
                    wl.running_slots, wl.plugin, wl.cms_name, wlst.name AS state
               FROM wmbs_location wl
                 INNER JOIN wmbs_location_pnns wls ON wls.location = wl.id
                 INNER JOIN wmbs_pnns wpnn ON wpnn.id = wls.pnn
                 INNER JOIN wmbs_location_state wlst ON wlst.id = wl.state
          """

    def execute(self, siteName=None, conn=None, transaction=False):
        if not siteName:
            results = self.dbi.processData(self.sql, conn=conn,
                                           transaction=transaction)
        else:
            sql = self.sql + " WHERE wl.site_name = :site"
            results = self.dbi.processData(sql, {'site': siteName},
                                       conn=conn, transaction=transaction)
        return self.format(results)

    def format(self, result):
        """
        Format the DB results in a plain list of dictionaries, with one
        dictionary for each site name, thus with a list of PNNs.
        :param result: DBResult object
        :return: a list of dictionaries
        """
        # first create a dictionary to make key look-up easier
        resp = {}
        for thisItem in DBFormatter.format(self, result):
            # note that each item has 7 columns returned from the database
            siteName = thisItem[0]
            resp.setdefault(siteName, dict())
            siteInfo = resp[siteName]
            siteInfo['site_name'] = siteName
            siteInfo.setdefault('pnn', [])
            if thisItem[1] not in siteInfo['pnn']:
                siteInfo['pnn'].append(thisItem[1])
            siteInfo['ce_name'] = thisItem[2]
            siteInfo['pending_slots'] = thisItem[3]
            siteInfo['running_slots'] = thisItem[4]
            siteInfo['plugin'] = thisItem[5]
            siteInfo['cms_name'] = thisItem[6]
            siteInfo['state'] = thisItem[7]
        # now return a flat list of dictionaries
        return list(resp.values())



#!/usr/bin/env python
# pylint: disable=W0212
# W0212 (protected-access): Access to protected names of a client class.
"""
_WMTask_

Object containing a set of executable Steps which form a template for a
set of jobs.

Equivalent of a WorkflowSpec in the ProdSystem.
"""
import json
from builtins import map, zip, str as newstr, bytes
from future.utils import viewitems

import logging
import os.path
import time

import WMCore.WMSpec.Steps.StepFactory as StepFactory
import WMCore.WMSpec.Utilities as SpecUtils
from WMCore.Configuration import ConfigSection
from WMCore.DataStructs.LumiList import LumiList
from WMCore.DataStructs.Workflow import Workflow as DataStructsWorkflow
from WMCore.Lexicon import lfnBase
from WMCore.WMSpec.ConfigSectionTree import ConfigSectionTree, TreeHelper
from WMCore.WMSpec.Steps.BuildMaster import BuildMaster
from WMCore.WMSpec.Steps.ExecuteMaster import ExecuteMaster
from WMCore.WMSpec.Steps.Template import CoreHelper
from WMCore.WMSpec.WMStep import WMStep, WMStepHelper


def getTaskFromStep(stepRef):
    """
    _getTaskFromStep_

    Traverse up the step tree until finding the first WMTask entry,
    return it wrapped in a WMTaskHelper

    """
    nodeData = stepRef
    if isinstance(stepRef, WMStepHelper):
        nodeData = stepRef.data

    taskNode = SpecUtils.findTaskAboveNode(nodeData)
    if taskNode is None:
        msg = "Unable to find Task containing step\n"
        # TODO: Replace with real exception class
        raise RuntimeError(msg)

    return WMTaskHelper(taskNode)


def buildLumiMask(runs, lumis):
    """
    Runs are saved in the spec as a list of integers.
    The lumi mask associated to each run is saved as a list of strings
    where each string is in a format like '1,4,23,45'

    The method convert these parameters in the corresponding lumiMask,
    e.g.:  runs=['3','4'], lumis=['1,4,23,45', '5,84,234,445'] => lumiMask = {'3':[[1,4],[23,45]],'4':[[5,84],[234,445]]}
    """

    if len(runs) != len(lumis):
        raise ValueError("runs and lumis must have same length")
    for lumi in lumis:
        if len(lumi.split(',')) % 2:
            raise ValueError("Needs an even number of lumi in each element of lumis list")

    lumiLists = [list(map(list, list(zip([int(y) for y in x.split(',')][::2], [int(y) for y in x.split(',')][1::2]))))
                 for x
                 in lumis]
    strRuns = [str(run) for run in runs]

    lumiMask = dict(list(zip(strRuns, lumiLists)))

    return lumiMask


class WMTaskHelper(TreeHelper):
    """
    _WMTaskHelper_

    Util wrapper containing tools & methods for manipulating the WMTask
    data object.
    """

    def __init__(self, wmTask):
        TreeHelper.__init__(self, wmTask)
        self.startTime = None
        self.endTime = None
        self.monitoring = None

    def addTask(self, taskName):
        """
        _addTask_

        Add a new task as a subtask with the name provided and
        return it wrapped in a TaskHelper

        """
        node = WMTaskHelper(WMTask(taskName))
        self.addNode(node)
        pName = "%s/%s" % (self.getPathName(), taskName)
        node.setPathName(pName)
        return node

    def taskIterator(self):
        """
        _taskIterator_

        return output of nodeIterator(self) wrapped in TaskHelper instance

        """
        for x in self.nodeIterator():
            yield WMTaskHelper(x)

    def childTaskIterator(self):
        """
        _childTaskIterator_

        Iterate over all the first generation child tasks.
        """
        for x in self.firstGenNodeChildIterator():
            yield WMTaskHelper(x)

    def setPathName(self, pathName):
        """
        _setPathName_

        Set the path name of the task within the workload
        Used internally when addin tasks to workloads or subtasks

        """
        self.data.pathName = pathName

    def getPathName(self):
        """
        _getPathName_

        get the path name of this task reflecting its
        structure within the workload and task tree

        """
        return self.data.pathName

    def name(self):
        """
        _name_

        Retrieve the name of this task.
        """
        return self.data._internal_name

    def listPathNames(self):
        """
        _listPathNames_

        """
        for t in self.taskIterator():
            yield t.getPathName()

    def listNames(self):
        """
        _listNames_
        Returns a generator with the name of all the children tasks
        """
        for t in self.taskIterator():
            yield t.name()

    def listChildNames(self):
        """
        _listChildNames_
        Return a list with the name of the first generation children tasks
        """
        names = []
        for t in self.childTaskIterator():
            names.append(t.name())
        return names

    def makeWorkflow(self):
        """
        _makeWorkflow_

        Create a WMBS compatible Workflow structure that represents this
        task and the information contained within it

        """
        workflow = DataStructsWorkflow()
        workflow.task = self.getPathName()
        return workflow

    def steps(self):
        """get WMStep structure"""
        if self.data.steps.topStepName is None:
            return None
        step = getattr(self.data.steps, self.data.steps.topStepName, None)
        return WMStepHelper(step)

    def getTopStepName(self):
        """
        _getTopStepName_

        Retrieve the name of the top step.
        """
        return self.data.steps.topStepName

    def setStep(self, wmStep):
        """set topStep to be the step instance provided"""
        stepData = wmStep
        if isinstance(wmStep, WMStepHelper):
            stepData = wmStep.data
            stepHelper = wmStep
        else:
            stepHelper = WMStepHelper(wmStep)

        stepName = stepHelper.name()
        stepHelper.setTopOfTree()
        setattr(self.data.steps, stepName, stepData)
        setattr(self.data.steps, "topStepName", stepName)
        return

    def listAllStepNames(self, cmsRunOnly=False):
        """
        _listAllStepNames_

        Get a list of all the step names contained in this task.
        """
        step = self.steps()
        if step:
            stepNames = step.allNodeNames()
            if cmsRunOnly:
                stepNames = [step for step in stepNames if step.startswith("cmsRun")]
            return stepNames
        else:
            return []

    def getStep(self, stepName):
        """get a particular step from the workflow"""
        if self.data.steps.topStepName is None:
            return None
        topStep = self.steps()
        return topStep.getStep(stepName)

    def makeStep(self, stepName):
        """
        _makeStep_

        create a new WMStep instance, install it as the top step and
        return the reference to the new step wrapped in a StepHelper

        """
        newStep = WMStep(stepName)
        self.setStep(newStep)
        return WMStepHelper(newStep)

    def applyTemplates(self):
        """
        _applyTemplates_

        For each step, load the appropriate template and install the default structure

        TODO: Exception handling

        """
        for step in self.steps().nodeIterator():
            stepType = step.stepType
            template = StepFactory.getStepTemplate(stepType)
            template(step)

    def getStepHelper(self, stepName):
        """
        _getStepHelper_

        Get the named step, look up its type specific helper and retrieve
        the step wrapped in the type based helper.

        """
        step = self.getStep(stepName)
        stepType = step.stepType()
        template = StepFactory.getStepTemplate(stepType)
        helper = template.helper(step.data)
        return helper

    def getOutputModulesForTask(self, cmsRunOnly=False):
        """
        _getOutputModulesForTask_

        Retrieve all the output modules in the given task.
        If cmsRunOnly is set to True, then return the output modules for
        cmsRun steps only.
        """
        outputModules = []
        for stepName in self.listAllStepNames(cmsRunOnly):
            outputModules.append(self.getOutputModulesForStep(stepName))
        return outputModules

    def getIgnoredOutputModulesForTask(self):
        """
        _getIgnoredOutputModulesForTask_

        Retrieve the ignored output modules in the given task.
        """
        ignoredOutputModules = []
        for stepName in self.listAllStepNames():
            stepHelper = self.getStepHelper(stepName)
            ignoredOutputModules.extend(stepHelper.getIgnoredOutputModules())
        return ignoredOutputModules

    def getOutputModulesForStep(self, stepName):
        """
        _getOutputModulesForStep_

        Retrieve all the output modules for the particular step.
        """
        step = self.getStep(stepName)

        if hasattr(step.data, "output"):
            if hasattr(step.data.output, "modules"):
                return step.data.output.modules

        return ConfigSection()

    def build(self, workingDir):
        """
        _build_

        Invoke the build process to create the job in the working dir provided

        """
        master = BuildMaster(workingDir)
        master(self)
        return

    def addEnvironmentVariables(self, envDict):
        """
        _addEnvironmentVariables_

        add a key = value style setting to the environment for this task and all
        its children
        """
        for key, value in viewitems(envDict):
            setattr(self.data.environment, key, value)
        for task in self.childTaskIterator():
            task.addEnvironmentVariables(envDict)
        return

    def setOverrideCatalog(self, tfcFile):
        """
        _setOverrideCatalog_

        Used for setting overrideCatalog option for each step in the task.
        """
        for step in self.steps().nodeIterator():
            step = CoreHelper(step)
            step.setOverrideCatalog(tfcFile)
        for task in self.childTaskIterator():
            task.setOverrideCatalog(tfcFile)
        return

    def getEnvironmentVariables(self):
        """
        _getEnvironmentVariables_

        Retrieve a dictionary with all environment variables defined for this task
        """
        return self.data.environment.dictionary_()

    def setupEnvironment(self):
        """
        _setupEnvironment_

        I don't know if this should go here.
        Setup the environment variables mandated in the WMTask
        """

        if not hasattr(self.data, 'environment'):
            # No environment to setup, pass
            return

        envDict = self.data.environment.dictionary_()

        for key in envDict:
            if str(envDict[key].__class__) == "<class 'WMCore.Configuration.ConfigSection'>":
                # At this point we do not support the
                # setting of sub-sections for environment variables
                continue
            else:
                os.environ[key] = envDict[key]

        return

    def execute(self, wmbsJob):
        """
        _execute_

        Invoke execution of the steps

        """
        self.startTime = time.time()
        self.setupEnvironment()
        master = ExecuteMaster()
        master(self, wmbsJob)
        self.endTime = time.time()
        return

    def setInputReference(self, stepRef, **extras):
        """
        _setInputReference_

        Add details to the input reference for the task providing
        input to this task.
        The reference is the step in the input task, plus
        any extra information.


        """
        stepId = SpecUtils.stepIdentifier(stepRef)
        setattr(self.data.input, "inputStep", stepId)
        for key, val in viewitems(extras):
            setattr(self.data.input, key, val)

        return

    def setInputStep(self, stepName):
        """
        _setInputStep_

        Set the name of the step used who's output is used as input for this
        task.
        """
        self.data.input.inputStep = stepName
        return

    def getInputStep(self):
        """
        _getInputStep_

        Retrieve the name of the input step, if there is one.
        """
        return getattr(self.data.input, "inputStep", None)

    def inputReference(self):
        """
        _inputReference_

        Get information about the input reference for this task.

        """
        return self.data.input

    def setFirstEventAndLumi(self, firstEvent, firstLumi):
        """
        _setFirstEventAndLumi_

        Set an arbitrary first event and first lumi
        Only used by production workflows
        """

        if not hasattr(self.data, "production"):
            self.data._section("production")
        setattr(self.data.production, "firstEvent", firstEvent)
        setattr(self.data.production, "firstLumi", firstLumi)

    def getFirstEvent(self):
        """
        _getFirstEvent_

        Get first event to produce for the task
        """
        if hasattr(self.data, "production"):
            if hasattr(self.data.production, "firstLumi"):
                return self.data.production.firstEvent
        return 1

    def getFirstLumi(self):
        """
        _getFirstLumi_

        Get first lumi to produce for the task
        """
        if hasattr(self.data, "production"):
            if hasattr(self.data.production, "firstLumi"):
                return self.data.production.firstLumi
        return 1

    def setSplittingParameters(self, **params):
        """
        _setSplittingParameters_

        Set the job splitting parameters.
        """
        for key, val in viewitems(params):
            setattr(self.data.input.splitting, key, val)

        return

    def setSplittingAlgorithm(self, algoName, **params):
        """
        _setSplittingAlgorithm_

        Set the splitting algorithm name and arguments.  Clear out any old
        splitting parameters while preserving the parameters for ACDC
        resubmission which are:
          collectionName, filesetName, couchURL, couchDB, owner, group

        This also needs to preserve the parameter we use to set the initial
        LFN counter, whether or not we merge across runs and the runWhitelist:
          initial_lfn_counter
          merge_across_runs
          runWhitelist

        Preserve parameters which can be set up at request creation and if not
        specified should remain unchanged, at the moment these are:
            include_parents
            lheInputFiles

        Also preserve the performance section.
        """
        setACDCParams = {}
        for paramName in ["collectionName", "filesetName", "couchURL",
                          "couchDB", "owner", "group", "initial_lfn_counter",
                          "merge_across_runs", "runWhitelist"]:
            if hasattr(self.data.input.splitting, paramName):
                setACDCParams[paramName] = getattr(self.data.input.splitting,
                                                   paramName)
        preservedParams = {}
        for paramName in ["lheInputFiles", "include_parents", "deterministicPileup"]:
            if hasattr(self.data.input.splitting, paramName):
                preservedParams[paramName] = getattr(self.data.input.splitting,
                                                     paramName)
        performanceConfig = getattr(self.data.input.splitting, "performance", None)

        delattr(self.data.input, "splitting")
        self.data.input.section_("splitting")
        self.data.input.splitting.section_("performance")

        setattr(self.data.input.splitting, "algorithm", algoName)
        self.setSplittingParameters(**preservedParams)
        self.setSplittingParameters(**params)
        self.setSplittingParameters(**setACDCParams)
        if performanceConfig is not None:
            self.data.input.splitting.performance = performanceConfig
        return

    def updateSplittingParameters(self, algoName, **params):
        """
        _updateSplittingAlgorithm_
        :param algoName: string Algorithm name
        :param params: splitting parameters
        :return:

        Only updates specific parameters in splitting Algorithm but doesn't remove the existing splitting parameters
        """
        performanceConfig = getattr(self.data.input.splitting, "performance", None)
        setattr(self.data.input.splitting, "algorithm", algoName)
        self.data.input.splitting.section_("performance")
        self.setSplittingParameters(**params)
        if performanceConfig is not None:
            self.data.input.splitting.performance = performanceConfig
        return

    def jobSplittingAlgorithm(self):
        """
        _jobSplittingAlgorithm_

        Retrieve the job splitting algorithm name.
        """
        return getattr(self.data.input.splitting, "algorithm", None)

    def jobSplittingParameters(self, performance=True):
        """
        _jobSplittingParameters_

        Retrieve the job splitting parameters.  This will combine the job
        splitting parameters specified in the spec with the site white list
        and black list as those are passed to the job splitting code.
        If required, also extract the performance parameters and pass them in the dict.
        """
        datadict = getattr(self.data.input, "splitting")
        if performance:
            splittingParams = datadict.dictionary_whole_tree_()
        else:
            splittingParams = datadict.dictionary_()
            if "performance" in splittingParams:
                del splittingParams['performance']
        splittingParams["siteWhitelist"] = self.siteWhitelist()
        splittingParams["siteBlacklist"] = self.siteBlacklist()
        splittingParams["trustSitelists"] = self.getTrustSitelists().get('trustlists')
        splittingParams["trustPUSitelists"] = self.getTrustSitelists().get('trustPUlists')

        if "runWhitelist" not in splittingParams and self.inputRunWhitelist() is not None:
            splittingParams["runWhitelist"] = self.inputRunWhitelist()
        if "runBlacklist" not in splittingParams and self.inputRunBlacklist() is not None:
            splittingParams["runBlacklist"] = self.inputRunBlacklist()

        return splittingParams

    def setJobResourceInformation(self, timePerEvent=None, sizePerEvent=None, memoryReq=None):
        """
        _setJobResourceInformation_

        Set the values to estimate the required computing resources for a job,
        the three key values are main memory usage, time per processing unit (e.g. time per event) and
        disk usage per processing unit (e.g. size per event).
        """
        if self.taskType() in ["Merge", "Cleanup", "LogCollect"]:
            # don't touch job requirements for these task types
            return

        performanceParams = getattr(self.data.input.splitting, "performance")

        timePerEvent = timePerEvent.get(self.name()) if isinstance(timePerEvent, dict) else timePerEvent
        sizePerEvent = sizePerEvent.get(self.name()) if isinstance(sizePerEvent, dict) else sizePerEvent
        memoryReq = memoryReq.get(self.name()) if isinstance(memoryReq, dict) else memoryReq

        if timePerEvent or getattr(performanceParams, "timePerEvent", None):
            performanceParams.timePerEvent = timePerEvent or getattr(performanceParams, "timePerEvent")
        if sizePerEvent or getattr(performanceParams, "sizePerEvent", None):
            performanceParams.sizePerEvent = sizePerEvent or getattr(performanceParams, "sizePerEvent")
        if memoryReq or getattr(performanceParams, "memoryRequirement", None):
            performanceParams.memoryRequirement = memoryReq or getattr(performanceParams, "memoryRequirement")
            # if we change memory requirements, then we must change MaxPSS as well
            self.setMaxPSS(performanceParams.memoryRequirement)

        return

    def addGenerator(self, generatorName, **settings):
        """
        _addGenerator_


        """
        if 'generators' not in self.data.listSections_():
            self.data.section_('generators')
        if generatorName not in self.data.generators.listSections_():
            self.data.generators.section_(generatorName)

        helper = TreeHelper(getattr(self.data.generators, generatorName))
        helper.addValue(settings)

        return

    def listGenerators(self):
        """
        _listGenerators_

        """
        generators = getattr(self.data, "generators", None)
        if generators is None:
            return []
        return generators.listSections_()

    def getGeneratorSettings(self, generatorName):
        """
        _getGeneratorSettings_

        Extract the settings from the generator fields
        """
        generators = getattr(self.data, "generators", None)
        if generators is None:
            return {}
        generator = getattr(generators, generatorName, None)
        if generator is None:
            return {}

        confValues = TreeHelper(generator)
        args = {}
        tempArgs = confValues.pythoniseDict(sections=False)
        for entry in tempArgs:
            args[entry.split('%s.' % generatorName)[1]] = tempArgs[entry]
        return args

    def addInputACDC(self, serverUrl, databaseName, collectionName,
                     filesetName):
        """
        _addInputACDC_

        Set the ACDC input information for this task.
        """
        self.data.input.section_("acdc")
        self.data.input.acdc.server = serverUrl
        self.data.input.acdc.database = databaseName
        self.data.input.acdc.collection = collectionName
        self.data.input.acdc.fileset = filesetName
        return

    def getInputACDC(self):
        """
        _getInputACDC_

        Retrieve the ACDC input configuration.
        """
        if not hasattr(self.data.input, "acdc"):
            return None

        return {"server": self.data.input.acdc.server,
                "collection": self.data.input.acdc.collection,
                "fileset": self.data.input.acdc.fileset,
                "database": self.data.input.acdc.database}

    def addInputDataset(self, **options):
        """
        _addInputDataset_

        Add details of an input dataset to this Task.
        This dataset will be used as input for the first step
        in the task

        options should contain at least:
          - name - dataset name
          - primary - primary dataset name
          - processed - processed dataset name
          - tier - data tier name

        optional args:
          - dbsurl - dbs url if not global
          - block_whitelist - list of whitelisted fileblocks
          - block_blacklist - list of blacklisted fileblocks
          - run_whitelist - list of whitelist runs
          - run_blacklist - list of blacklist runs
        """
        self.data.input.section_("dataset")
        self.data.input.dataset.name = None
        self.data.input.dataset.dbsurl = None
        self.data.input.dataset.section_("blocks")
        self.data.input.dataset.blocks.whitelist = []
        self.data.input.dataset.blocks.blacklist = []
        self.data.input.dataset.section_("runs")
        self.data.input.dataset.runs.whitelist = []
        self.data.input.dataset.runs.blacklist = []

        try:
            self.data.input.dataset.primary = options.pop('primary')
            self.data.input.dataset.processed = options.pop('processed')
            self.data.input.dataset.tier = options.pop('tier')
        except KeyError:
            raise RuntimeError("Primary, Processed and Tier must be set")

        for opt, arg in viewitems(options):
            if opt == 'block_blacklist':
                self.setInputBlockBlacklist(arg)
            elif opt == 'block_whitelist':
                self.setInputBlockWhitelist(arg)
            elif opt == 'dbsurl':
                self.data.input.dataset.dbsurl = arg
            elif opt == "run_whitelist":
                self.setInputRunWhitelist(arg)
            elif opt == "run_blacklist":
                self.setInputRunBlacklist(arg)
            else:
                setattr(self.data.input.dataset, opt, arg)

        return

    def setInputBlockWhitelist(self, blockWhitelist):
        """
        _setInputBlockWhitelist_

        Set the block white list for the input dataset.  This must be called
        after setInputDataset().
        """
        self.data.input.dataset.blocks.whitelist = blockWhitelist
        return

    def inputBlockWhitelist(self):
        """
        _inputBlockWhitelist_

        Retrieve the block white list for the input dataset if it exists, none
        otherwise.
        """
        if hasattr(self.data.input, "dataset"):
            return self.data.input.dataset.blocks.whitelist
        return None

    def setInputBlockBlacklist(self, blockBlacklist):
        """
        _setInputBlockBlacklist_

        Set the block black list for the input dataset.  This must be called
        after setInputDataset().
        """
        self.data.input.dataset.blocks.blacklist = blockBlacklist
        return

    def inputBlockBlacklist(self):
        """
        _inputBlockBlacklist_

        Retrieve the block black list for the input dataset if it exsits, none
        otherwise.
        """
        if hasattr(self.data.input, "dataset"):
            return self.data.input.dataset.blocks.blacklist
        return None

    def setInputRunWhitelist(self, runWhitelist):
        """
        _setInputRunWhitelist_

        Set the run white list for the input dataset.  This must be called
        after setInputDataset().
        """
        self.data.input.dataset.runs.whitelist = runWhitelist
        return

    def inputRunWhitelist(self):
        """
        _inputRunWhitelist_

        Retrieve the run white list for the input dataset if it exists, none
        otherwise.
        """
        if hasattr(self.data.input, "dataset"):
            return self.data.input.dataset.runs.whitelist
        return None

    def setInputRunBlacklist(self, runBlacklist):
        """
        _setInputRunBlacklist_

        Set the run black list for the input dataset.  This must be called
        after setInputDataset().
        """
        self.data.input.dataset.runs.blacklist = runBlacklist
        return

    def inputRunBlacklist(self):
        """
        _inputRunBlacklist_

        Retrieve the run black list for the input dataset if it exists, none
        otherwise.
        """
        if hasattr(self.data.input, "dataset"):
            return self.data.input.dataset.runs.blacklist
        return None

    def addProduction(self, **options):
        """
        _addProduction_

        Add details of production job related information.

        options should contain at least:
        TODO: Not sure what is necessary data ask Dave
        optional
        - totalevents - total events in dataset

        """
        if not hasattr(self.data, "production"):
            self.data.section_("production")

        for opt, arg in viewitems(options):
            setattr(self.data.production, opt, arg)

    def inputDataset(self):
        """
        _inputDataset_

        Get the input.dataset structure from this task

        """
        return getattr(self.data.input, "dataset", None)

    def getInputDatasetPath(self):
        """
        _getInputDatasetPath_

        Get the input dataset path because it's useful
        """

        if hasattr(self.data.input, 'dataset'):
            return getattr(self.data.input.dataset, 'name', None)

        return None

    def setInputPileupDatasets(self, dsetName):
        """
        _setInputPileupDatasets_

        Create a list of pileup datasets to be used by this task (possible
        multiple CMSSW steps)
        """
        self.data.input.section_("pileup")
        if not hasattr(self.data.input.pileup, "datasets"):
            self.data.input.pileup.datasets = []

        if isinstance(dsetName, list):
            self.data.input.pileup.datasets.extend(dsetName)
        elif isinstance(dsetName, (newstr, bytes)):
            self.data.input.pileup.datasets.append(dsetName)
        else:
            raise ValueError("Pileup dataset must be either a list or a string (unicode or bytes)")
        # make the list unique
        self.data.input.pileup.datasets = list(set(self.data.input.pileup.datasets))

    def getInputPileupDatasets(self):
        """
        _getInputPileupDatasets_

        Get a list of the input pileup dataset name(s) for this task.
        """
        if hasattr(self.data.input, 'pileup'):
            return getattr(self.data.input.pileup, 'datasets', [])
        return []

    def siteWhitelist(self):
        """
        _siteWhitelist_

        Accessor for the site white list for the task.
        """
        return self.data.constraints.sites.whitelist

    def setSiteWhitelist(self, siteWhitelist):
        """
        _setSiteWhitelist_

        Set the set white list for the task.
        """
        self.data.constraints.sites.whitelist = siteWhitelist
        return

    def siteBlacklist(self):
        """
        _siteBlacklist_

        Accessor for the site white list for the task.
        """
        return self.data.constraints.sites.blacklist

    def setSiteBlacklist(self, siteBlacklist):
        """
        _setSiteBlacklist_

        Set the site black list for the task.
        """
        self.data.constraints.sites.blacklist = siteBlacklist
        return

    def getTrustSitelists(self):
        """
        _getTrustSitelists_

        Get the input and pileup flag for 'trust site lists' in the task.
        """
        # handle backward compatibility for the request which doesn't contain trustPUlists
        return {'trustlists': getattr(self.data.constraints.sites, 'trustlists', False),
                'trustPUlists': getattr(self.data.constraints.sites, 'trustPUlists', False)}

    def setTrustSitelists(self, trustSitelists, trustPUSitelists):
        """
        _setTrustSitelists_

        Set the input and the pileup flags for 'trust site lists' in the task.
        """
        self.data.constraints.sites.trustlists = trustSitelists
        self.data.constraints.sites.trustPUlists = trustPUSitelists
        return

    def listOutputDatasetsAndModules(self):
        """
        _listOutputDatasetsAndModules_

        Get the output datasets per output module for this task
        """
        outputDatasets = []
        for stepName in self.listAllStepNames():
            stepHelper = self.getStepHelper(stepName)

            if not getattr(stepHelper.data.output, "keep", True):
                continue

            if stepHelper.stepType() == "CMSSW":
                for outputModuleName in stepHelper.listOutputModules():
                    outputModule = stepHelper.getOutputModule(outputModuleName)
                    outputDataset = "/%s/%s/%s" % (outputModule.primaryDataset,
                                                   outputModule.processedDataset,
                                                   outputModule.dataTier)
                    outputDatasets.append({"outputModule": outputModuleName,
                                           "outputDataset": outputDataset})

        return outputDatasets

    def setSubscriptionInformation(self, custodialSites=None, nonCustodialSites=None,
                                   priority="Low", primaryDataset=None,
                                   useSkim=False, isSkim=False,
                                   dataTier=None, deleteFromSource=False,
                                   datasetLifetime=None):
        """
        _setSubscriptionsInformation_

        Set the subscription information for this task's datasets
        The subscriptions information is structured as follows:
        data.subscriptions.outputSubs is a list with the output section names (1 per dataset)
        data.subscriptions.<outputSection>.dataset
        data.subscriptions.<outputSection>.outputModule
        data.subscriptions.<outputSection>.custodialSites
        data.subscriptions.<outputSection>.nonCustodialSites
        data.subscriptions.<outputSection>.priority

        The filters arguments allow to define a dataTier and primaryDataset. Only datasets
        matching those values will be configured.
        """
        custodialSites = custodialSites or []
        nonCustodialSites = nonCustodialSites or []

        if not hasattr(self.data, "subscriptions"):
            self.data.section_("subscriptions")
            self.data.subscriptions.outputSubs = []

        outputDatasets = self.listOutputDatasetsAndModules()

        for entry in enumerate(outputDatasets, start=1):
            subSectionName = "output%s" % entry[0]
            outputDataset = entry[1]["outputDataset"]
            outputModule = entry[1]["outputModule"]

            dsSplit = outputDataset.split('/')
            primDs = dsSplit[1]
            tier = dsSplit[3]
            procDsSplit = dsSplit[2].split('-')
            skim = (len(procDsSplit) == 4)

            if primaryDataset and primDs != primaryDataset:
                continue
            if useSkim and isSkim != skim:
                continue
            if dataTier and tier != dataTier:
                continue

            self.data.subscriptions.outputSubs.append(subSectionName)
            outputSection = self.data.subscriptions.section_(subSectionName)
            outputSection.dataset = outputDataset
            outputSection.outputModule = outputModule
            outputSection.custodialSites = custodialSites
            outputSection.nonCustodialSites = nonCustodialSites
            outputSection.priority = priority
            outputSection.deleteFromSource = deleteFromSource
            outputSection.datasetLifetime = datasetLifetime

        return

    def getSubscriptionInformation(self):
        """
        _getSubscriptionInformation_

        Get the subscription configuration for the task
        return a dictionary with the following structure
        {<dataset> : {CustodialSites : [],
                      NonCustodialSites : [],
                      Priority : "Low"
                     }
        }
        """
        if not hasattr(self.data, "subscriptions"):
            return {}

        subKeyName = 'outputSubs'

        subInformation = {}
        for outputSub in getattr(self.data.subscriptions, subKeyName):
            outputSection = getattr(self.data.subscriptions, outputSub)
            dataset = outputSection.dataset

            subInformation[dataset] = {"CustodialSites": outputSection.custodialSites,
                                       "NonCustodialSites": outputSection.nonCustodialSites,
                                       "Priority": outputSection.priority,
                                       # These might not be present in all specs
                                       "DeleteFromSource": getattr(outputSection, "deleteFromSource", False),
                                       # Spec assigned for T0 ContainerRules
                                       "DatasetLifetime": getattr(outputSection, "datasetLifetime", 0)}
        return subInformation

    def parentProcessingFlag(self):
        """
        _parentProcessingFlag_

        accessor for parentProcessing information (two file input)
        """
        return self.jobSplittingParameters().get("include_parents", False)

    def totalEvents(self):
        """
        _totalEvents_

        accessor for total events in the given dataset
        """
        # TODO: save the total events for  the production job
        return int(self.data.production.totalEvents)
        # return self.data.input.dataset.totalEvents

    def dbsUrl(self):
        """
        _dbsUrl_
        if local dbs url is set for the task, return it
        otherwise return None
        """
        if getattr(self.data.input, "dataset", False):
            return getattr(self.data.input.dataset, "dbsurl", None)
        else:
            return None

    def setTaskType(self, taskType):
        """
        Set the type field of this task
        """
        self.data.taskType = taskType

    def taskType(self):
        """
        _taskType_

        Get the task Type setting
        """
        return self.data.taskType

    def completeTask(self, jobLocation, reportName):
        """
        _completeTask_

        Combine all the logs from all the steps in the task to a single log

        If necessary, output to Dashboard
        """
        from WMCore.FwkJobReport.Report import Report

        finalReport = Report()
        # We left the master report at the pilot scratch area level
        testPath = os.path.join(jobLocation, '../../', reportName)
        logging.info("Looking for master report at %s", testPath)
        if os.path.exists(testPath):
            logging.info("  found it!")
            # If a report already exists, we load it and
            # append our steps to it
            finalReport.load(testPath)
        taskSteps = self.listAllStepNames()
        for taskStep in taskSteps:
            reportPath = os.path.join(jobLocation, taskStep, "Report.pkl")
            logging.info("Looking for a taskStep report at %s", reportPath)
            if os.path.isfile(reportPath):
                logging.info("  found it!")
                stepReport = Report()
                stepReport.unpersist(reportPath, taskStep)
                finalReport.setStep(taskStep, stepReport.retrieveStep(taskStep))
                logURL = stepReport.getLogURL()
                if logURL:
                    finalReport.setLogURL(logURL)
            else:
                msg = "  failed to find it."
                msg += "Files in the directory are:\n%s" % os.listdir(os.path.join(jobLocation, taskStep))
                logging.error(msg)
                # Then we have a missing report
                # This should raise an alarm bell, as per Steve's request
                # TODO: Change error code
                finalReport.addStep(reportname=taskStep, status=1)
                finalReport.addError(stepName=taskStep, exitCode=99996, errorType="ReportManipulatingError",
                                     errorDetails="Failed to find a step report for %s!" % taskStep)

        finalReport.data.completed = True
        finalReport.persist(reportName)

        return finalReport

    def taskLogBaseLFN(self):
        """
        _taskLogBaseLFN_

        Get the base LFN for the task's log archive file.
        """
        return getattr(self.data, "logBaseLFN", "/store/temp/WMAgent/unmerged")

    def setTaskLogBaseLFN(self, logBaseLFN):
        """
        _setTaskLogBaseLFN_

        Set the base LFN for the task's log archive file.
        """
        self.data.logBaseLFN = logBaseLFN
        return

    def addNotification(self, target):
        """
        _addNotification_

        Add a target to be notified on workflow completion
        """

        self.data.notifications.targets.append(target)
        return

    def getNotifications(self):
        """
        _getNotifications_

        Get all targets for notification at workflow completion
        """

        return self.data.notifications.targets

    def _setPerformanceMonitorConfig(self):
        """
        if config section for the PerformanceMonitor. If not set, it will set one
        """
        if self.monitoring is not None:
            return

        self.monitoring = self.data.section_("watchdog")
        if not hasattr(self.data.watchdog, 'monitors'):
            self.data.watchdog.monitors = []
        if 'PerformanceMonitor' not in self.monitoring.monitors:
            self.monitoring.monitors.append('PerformanceMonitor')
            self.monitoring.section_("PerformanceMonitor")
        return

    def setMaxPSS(self, maxPSS):
        """
        _setMaxPSS_

        Set MaxPSS performance monitoring for this task.
        :param maxPSS: maximum Proportional Set Size (PSS) memory consumption in MiB
        """
        if self.taskType() in ["Merge", "Cleanup", "LogCollect"]:
            # keep the default settings (from StdBase) for these task types
            return

        if isinstance(maxPSS, dict):
            maxPSS = maxPSS.get(self.name(), None)

        if maxPSS:
            self._setPerformanceMonitorConfig()
            self.monitoring.PerformanceMonitor.maxPSS = int(maxPSS)
            for task in self.childTaskIterator():
                task.setMaxPSS(maxPSS)
        return

    def setPerformanceMonitor(self, softTimeout=None, gracePeriod=None):
        """
        _setPerformanceMonitor_

        Set/Update the performance monitor options for the task
        """
        # make sure there is a PerformanceMonitor section in the task
        self._setPerformanceMonitorConfig()

        if softTimeout:
            self.monitoring.PerformanceMonitor.softTimeout = int(softTimeout)
            if gracePeriod:
                self.monitoring.PerformanceMonitor.hardTimeout = int(softTimeout + gracePeriod)

        return

    def getSwVersion(self, allSteps=False):
        """
        _getSwVersion_

        Get the CMSSW version for the first CMSSW step in this task.
        :param allSteps: set it to True to retrieve a list of CMSSW releases
         used in this task
        :return: a string with the release name or a list of releases if allSteps is True.
        """
        versions = []
        for stepName in self.listAllStepNames():
            stepHelper = self.getStepHelper(stepName)
            if stepHelper.stepType() in ["CMSSW", "LogCollect"]:
                if not allSteps:
                    return stepHelper.getCMSSWVersion()
                else:
                    versions.append(stepHelper.getCMSSWVersion())
        return versions

    def getScramArch(self, allSteps=False):
        """
        _getScramArch_

        Get the scram architecture for the first CMSSW step of workload.
        Set allSteps to true to retrieve all the scramArchs used in this task.
        """
        scrams = []
        for stepName in self.listAllStepNames():
            stepHelper = self.getStepHelper(stepName)
            if stepHelper.stepType() in ["CMSSW", "LogCollect"]:
                if not allSteps:
                    return stepHelper.getScramArch()
                else:
                    scrams.append(stepHelper.getScramArch())
        return scrams

    def setPrimarySubType(self, subType):
        """
        _setPrimarySubType_

        Set the subType that should be used by WorkQueue for the
        primary subscription
        """

        self.data.parameters.primarySubType = subType
        return

    def getPrimarySubType(self):
        """
        _getPrimarySubType_

        Retrieve the primary subType
        If not available, use the taskType
        """

        return getattr(self.data.parameters, 'primarySubType',
                       self.taskType())

    def getConfigCacheIDs(self):
        """
        _getConfigCacheIDs_

        Search constituent steps for ConfigCacheID
        """

        IDs = []
        for stepName in self.listAllStepNames():
            stepHelper = self.getStepHelper(stepName)
            ID = stepHelper.getConfigCacheID()
            if ID:
                IDs.append(ID)
        return IDs

    def getPhysicsTaskType(self):
        """
        Return the physics Task type
        :return: str
        """
        return getattr(self.data, 'physicsTaskType', None)

    def getStepPhysicsTypes(self):
        """
        Get the physics types for all cmsRun steps
        :return: list
        """
        types = []
        for stepName in self.listAllStepNames(cmsRunOnly=True):
           stepHelper = self.getStepHelper(stepName)
           stepType = stepHelper.getPhysicsType()
           if stepType:
               types.append(stepType)
           else:
               types.append("UNKNOWN")
        return types

    def setPhysicsTaskType(self):
        """
        Set the physics task type based on the steps parameter
        for a step.
        We basically expand the standard "Production/Processing" taskType to
        detail the process better. 
        For MC: "Production" or "Processing" to ["GENSIM", "GEN", "DIGI", "RECO", "DIGIRECO", "MINIAOD"]
        For Data: "Processing" to "Dataprocessing"
        :return: str
        """
        physicsTaskType = "UNKNOWN"
        if self.taskType() in ("Processing", "Production"):
            stepTypes = self.getStepPhysicsTypes()
            if "DataProcessing" in stepTypes:
                # For data, return a single DataProcessing step
                physicsTaskType = "DataProcessing"
            else:
                # For MC, join all physics steps
                physicsTaskType = ",".join(stepTypes)
        else:
            # Other task types are not physics types
            physicsTaskType = None

        self.data.physicsTaskType = physicsTaskType

    def setProcessingVersion(self, procVer, parentProcessingVersion=0, stepChainMap=False):
        """
        _setProcessingVersion_

        Set the task processing version
        """
        if isinstance(procVer, dict) and stepChainMap:
            taskProcVer = self._getStepValue(procVer, parentProcessingVersion)
            self._setStepProperty("ProcessingVersion", procVer, stepChainMap)
        elif isinstance(procVer, dict):
            taskProcVer = procVer.get(self.name(), parentProcessingVersion)
            if taskProcVer is None:
                for taskname in procVer:
                    if taskname in self.name():
                        taskProcVer = procVer[taskname]
        else:
            taskProcVer = procVer

        self.data.parameters.processingVersion = int(taskProcVer)
        for task in self.childTaskIterator():
            task.setProcessingVersion(procVer, taskProcVer, stepChainMap)
        return

    def getProcessingVersion(self):
        """
        _getProcessingVersion_

        Get the task processing version
        """
        return getattr(self.data.parameters, 'processingVersion', 0)

    def setProcessingString(self, procString, parentProcessingString=None, stepChainMap=False):
        """
        _setProcessingString_

        Set the task processing string
        """
        if isinstance(procString, dict) and stepChainMap:
            taskProcString = self._getStepValue(procString, parentProcessingString)
            self._setStepProperty("ProcessingString", procString, stepChainMap)
        elif isinstance(procString, dict):
            taskProcString = procString.get(self.name(), parentProcessingString)
            if taskProcString is None:
                for taskname in procString:
                    if taskname in self.name():
                        taskProcString = procString[taskname]
        else:
            taskProcString = procString

        self.data.parameters.processingString = taskProcString

        for task in self.childTaskIterator():
            task.setProcessingString(procString, taskProcString, stepChainMap)
        return

    def getProcessingString(self):
        """
        _getProcessingString_

        Get the task processing string
        """
        return getattr(self.data.parameters, 'processingString', None)

    def getCMSSWVersionsWithMergeTask(self):
        """
        _getCMSSWVersionsWithMergeTask_

        Get the all the cmssw versions for this task plus first generation merge task cmssw version.
        This will be used to validate and check in the script.
        Merge cmssw version should be the same as processing version
        """
        versions = set()
        for stepName in self.listAllStepNames():

            stepHelper = self.getStepHelper(stepName)
            if stepHelper.stepType() != "CMSSW":
                continue
            version = stepHelper.getCMSSWVersion()
            versions.add(version)

        for task in self.childTaskIterator():
            if task.taskType() == "Merge":
                for stepName in task.listAllStepNames():

                    stepHelper = task.getStepHelper(stepName)
                    if stepHelper.stepType() != "CMSSW":
                        continue
                    version = stepHelper.getCMSSWVersion()
                    versions.add(version)

        return versions

    def setNumberOfCores(self, cores, nStreams):
        """
        _setNumberOfCores_

        Set number of cores and event streams for each CMSSW step in this task and its children
        """
        if self.taskType() in ["Merge", "Harvesting", "Cleanup", "LogCollect"]:
            return

        if isinstance(cores, dict):
            taskCores = cores.get(self.name())
        else:
            taskCores = cores

        if isinstance(nStreams, dict):
            taskStreams = nStreams.get(self.name(), 0)
        else:
            taskStreams = nStreams

        if taskCores:
            for stepName in self.listAllStepNames():
                stepHelper = self.getStepHelper(stepName)
                if stepHelper.stepType() == "CMSSW":
                    stepHelper.setNumberOfCores(taskCores, taskStreams)

        for task in self.childTaskIterator():
            task.setNumberOfCores(cores, nStreams)

        return

    def getNumberOfCores(self):
        """
        Retrieves the number of cores for this task.
        If it's a multi-step task, it returns only the greatest value
        :return: an integer with the number of cores required by this task
        """
        maxCores = 1
        for stepName in self.listAllStepNames():
            stepHelper = self.getStep(stepName)
            maxCores = max(maxCores, stepHelper.getNumberOfCores())
        return maxCores

    def setTaskGPUSettings(self, requiresGPU, gpuParams):
        """
        Setter method for the GPU settings, applied to this Task object and
        all underneath CMSSW type step object.
        :param requiresGPU: string defining whether GPUs are needed. For TaskChains, it
            could be a dictionary key'ed by the taskname.
        :param gpuParams: GPU settings. A JSON encoded object, from either a None object
            or a dictionary. For TaskChains, it could be a dictionary key'ed by the taskname
        :return: nothing, the workload spec is updated in place.
        """
        # these job types shall not have these settings
        if self.taskType() in ["Merge", "Harvesting", "Cleanup", "LogCollect"]:
            return

        # default values come from StdBase
        if isinstance(requiresGPU, dict):
            thisTaskGPU = requiresGPU.get(self.name(), "forbidden")
        else:
            thisTaskGPU = requiresGPU

        decodedGpuParams = json.loads(gpuParams)
        if self.name() in decodedGpuParams:
            thisTaskGPUParams = decodedGpuParams[self.name()]
        else:
            thisTaskGPUParams = decodedGpuParams

        for stepName in self.listAllStepNames():
            stepHelper = self.getStepHelper(stepName)
            if stepHelper.stepType() == "CMSSW":
                stepHelper.setGPUSettings(thisTaskGPU, thisTaskGPUParams)

        for task in self.childTaskIterator():
            task.setTaskGPUSettings(requiresGPU, gpuParams)

        return

    def getRequiresGPU(self):
        """
        Return whether this task is supposed to use GPUs or not.
        If it's a multi-step task, decision follows this order:
          1. "required"
          2. "optional"
          3. "forbidden"
        :return: a string (default to "forbidden")
        """
        requiresGPU = set(["forbidden"])
        for stepName in self.listAllStepNames():
            stepHelper = self.getStep(stepName)
            if stepHelper.stepType() == "CMSSW" and stepHelper.getGPURequired():
                requiresGPU.add(stepHelper.getGPURequired())

        # now decide what value has higher weight
        if len(requiresGPU) == 1:
            return requiresGPU.pop()
        elif "required" in requiresGPU:
            return "required"
        elif "optional" in requiresGPU:
            return "optional"
        else:
            return "forbidden"

    def getGPURequirements(self):
        """
        Return the GPU requirements for this task.
        If it's a multi-step task, the first step with a meaningful
        dictionary value will be returned
        :return: a dictionary with the GPU requirements for this task
        """
        gpuRequirements = {}
        for stepName in sorted(self.listAllStepNames()):
            stepHelper = self.getStep(stepName)
            if stepHelper.stepType() == "CMSSW" and stepHelper.getGPURequirements():
                return stepHelper.getGPURequirements()
        return gpuRequirements

    def _getStepValue(self, keyDict, defaultValue):
        """
        __getStepValue_

        Maps this taskName - in somehow a hacky way - to a 'StepName' value
        that should exist in a StepChain request. Used only on tasks that have
        output module
        :param keyDict: a dict with either AcqEra/ProcStr/ProcVer key/value pairs,
        where the key corresponds to the StepName
        """
        if self.taskType() == "Merge":
            extractedTaskName = self.name().split("Merge")[0]
            value = keyDict.get(extractedTaskName)
        elif self.taskType() in ["Production", "Processing"]:
            value = keyDict.get(self.name())
        else:
            value = defaultValue

        return value

    def _setStepProperty(self, propertyName, propertyDict, stepMap):
        """
        For StepChain workloads, we also need to set AcqEra/ProcStr/ProcVer
        at the WMStep level, such that we can properly map different cmsRun
        steps - within the same task - to different meta data information.
        :param propertyName: the name of the property to set at step level
        :param propertyDict: a dictionary mapping StepName to its value
        :param stepMap: map between step name, step number and cmsRun number,
                        same as returned from the workload getStepMapping
        """
        propMethodMap = {"AcquisitionEra": "setAcqEra",
                         "ProcessingString": "setProcStr",
                         "ProcessingVersion": "setProcStr"}

        if self.taskType() not in ["Production", "Processing"]:
            # then there is no need to set anything, single cmsRun step at most
            return

        for stepName, stepValues in viewitems(stepMap):
            cmsRunNum = stepValues[1]
            stepHelper = self.getStepHelper(cmsRunNum)
            callableMethod = getattr(stepHelper, propMethodMap[propertyName])
            callableMethod(propertyDict[stepName])

    def setAcquisitionEra(self, era, parentAcquisitionEra=None, stepChainMap=False):
        """
        _setAcquistionEra_

        Set the task acquisition era
        """

        if isinstance(era, dict) and stepChainMap:
            taskEra = self._getStepValue(era, parentAcquisitionEra)
            self._setStepProperty("AcquisitionEra", era, stepChainMap)
        elif isinstance(era, dict):
            taskEra = era.get(self.name(), parentAcquisitionEra)
            if taskEra is None:
                # We cannot properly set AcqEra for ACDC of TaskChain Merge
                # failures, so we should look up for a similar taskname in
                # the acqera dict passed from the requestor
                for taskname in era:
                    if taskname in self.name():
                        taskEra = era[taskname]
        else:
            taskEra = era

        self.data.parameters.acquisitionEra = taskEra

        for task in self.childTaskIterator():
            task.setAcquisitionEra(era, taskEra, stepChainMap)
        return

    def getAcquisitionEra(self):
        """
        _getAcquisitionEra_

        Get the task acquisition era.
        """
        return getattr(self.data.parameters, 'acquisitionEra', None)

    def setCampaignName(self, campaign):
        """
        Set the campaign name of this task
        :param campaign: str, name of the campaign to be defined
        """
        self.data.campaignName = campaign

    def getCampaignName(self):
        """
        Get the task campaign name
        :return: str, campaign name value
        """
        return getattr(self.data, 'campaignName', None)

    def setLumiMask(self, lumiMask=None, override=True):
        """
        Attach the given LumiMask to the task
        At this point the lumi mask is just the compactList dict not the LumiList object
        """

        if not lumiMask:
            return

        runs = getattr(self.data.input.splitting, 'runs', None)
        lumis = getattr(self.data.input.splitting, 'lumis', None)
        if not override and runs and lumis:  # Unless instructed, don't overwrite runs and lumis which may be there from a task already
            return

        runs = []
        lumis = []
        for run, runLumis in viewitems(lumiMask):
            runs.append(int(run))
            lumiList = []
            for lumi in runLumis:
                lumiList.extend([str(l) for l in lumi])
            lumis.append(','.join(lumiList))

        self.data.input.splitting.runs = runs
        self.data.input.splitting.lumis = lumis

        for task in self.childTaskIterator():
            task.setLumiMask(lumiMask, override)

        return

    def getLumiMask(self):
        """
            return the lumi mask
        """
        runs = getattr(self.data.input.splitting, 'runs', None)
        lumis = getattr(self.data.input.splitting, 'lumis', None)
        if runs and lumis:
            return LumiList(wmagentFormat=(runs, lumis))

        return {}

    def _propMethodMap(self):
        """
        internal mapping methop which maps which method need to be call for each
        property.
        For now only contains properties which updates in assignment stage.
        """
        propMap = {"ProcessingVersion": self.setProcessingVersion,
                   "AcquisitionEra": self.setAcquisitionEra,
                   "ProcessingString": self.setProcessingString
                   }
        return propMap

    def setProperties(self, properties):
        """
        set task properties (only for assignment stage but make it more general)
        """
        for prop, value in viewitems(properties):
            self._propMethodMap()[prop](value)

    def deleteChild(self, childName):
        """
        _deleteChild_

        Remove the child task from the tree, if it exists
        """
        self.deleteNode(childName)

    def setPrepID(self, prepID):
        """
        _setPrepID_

        Set the prepID to for all the tasks below
        """
        # if prepID doesn exist set it, if exist ignore.
        if not self.getPrepID() and prepID:
            self.data.prepID = prepID

        prepID = self.getPrepID()
        # set child prepid
        if prepID:
            for task in self.childTaskIterator():
                task.setPrepID(prepID)

    def getPrepID(self):
        """
        _getPrepID_

        Get the prepID for the workflow
        """
        return getattr(self.data, 'prepID', None)

    def setLFNBase(self, mergedLFNBase, unmergedLFNBase):
        """
        _setLFNBase_

        Set the merged and unmerged base LFNs for all tasks.
        """
        self.data.mergedLFNBase = mergedLFNBase
        self.data.unmergedLFNBase = unmergedLFNBase
        for task in self.childTaskIterator():
            task.setLFNBase(mergedLFNBase, unmergedLFNBase)

        return

    def _getLFNBase(self):
        """
        private method getting lfn base.
        lfn base should be set by workflow
        """
        return (getattr(self.data, 'mergedLFNBase', "/store/data"),
                getattr(self.data, 'unmergedLFNBase', "/store/unmerged"))

    def _getKeyValue(self, keyname, stepname, values):
        if keyname not in values:
            return
        elif isinstance(values[keyname], (newstr, bytes)):
            return values[keyname]
        elif isinstance(values[keyname], dict):
            return values[keyname].get(stepname)

    def _updateLFNsStepChain(self, stepName, dictValues, stepMapping):
        """
        __updateLFNsStepChain_

        Helper function needed for a proper StepChain LFN/ProcessedDataset handling

        :param stepName: is the cmsRun name (cmsRun1, cmsRun2, ...)
        :param dictValues: part of the arguments provided during assignment
        :param stepMapping: built during StepChain creation
        :return: a single string for each of those 3 properties
        """
        reqStepName = None
        for reqStep, values in viewitems(stepMapping):
            if stepName == values[1]:
                reqStepName = reqStep
        if not reqStepName:
            # I have no idea which cmsRun is that...
            return None, None, None

        era = self._getKeyValue('AcquisitionEra', reqStepName, dictValues)
        if not era:
            era = self.getAcquisitionEra()
        procstr = self._getKeyValue('ProcessingString', reqStepName, dictValues)
        if not procstr:
            procstr = self.getProcessingString()
        procver = self._getKeyValue('ProcessingVersion', reqStepName, dictValues)
        if not procver:
            procver = self.getProcessingVersion()

        return era, procstr, procver

    def updateLFNsAndDatasets(self, runNumber=None, dictValues=None, stepMapping=None):
        """
        _updateLFNsAndDatasets_

        Update all the output LFNs and data names for all tasks in the workflow.
        This needs to be called after updating the acquisition era, processing
        version or merged/unmerged lfn base.
        """
        mergedLFNBase, unmergedLFNBase = self._getLFNBase()
        taskType = self.taskType()

        for stepName in self.listAllStepNames():
            stepHelper = self.getStepHelper(stepName)

            if stepHelper.stepType() == "CMSSW":
                if dictValues and stepMapping:
                    # if it's a StepChain, then cast a dark spell on it
                    acqera, procstr, procver = self._updateLFNsStepChain(stepName, dictValues, stepMapping)
                else:
                    acqera = self.getAcquisitionEra()
                    procstr = self.getProcessingString()
                    procver = self.getProcessingVersion()

                for outputModuleName in stepHelper.listOutputModules():
                    outputModule = stepHelper.getOutputModule(outputModuleName)
                    filterName = getattr(outputModule, "filterName", None)

                    if procstr:
                        processingEra = "%s-v%i" % (procstr, procver)
                    else:
                        processingEra = "v%i" % procver
                    if filterName:
                        processedDataset = "%s-%s-%s" % (acqera, filterName, processingEra)
                        processingString = "%s-%s" % (filterName, processingEra)
                    else:
                        processedDataset = "%s-%s" % (acqera, processingEra)
                        processingString = processingEra

                    unmergedLFN = "%s/%s/%s/%s/%s" % (unmergedLFNBase,
                                                      acqera,
                                                      getattr(outputModule, "primaryDataset"),
                                                      getattr(outputModule, "dataTier"),
                                                      processingString)
                    mergedLFN = "%s/%s/%s/%s/%s" % (mergedLFNBase,
                                                    acqera,
                                                    getattr(outputModule, "primaryDataset"),
                                                    getattr(outputModule, "dataTier"),
                                                    processingString)

                    if runNumber is not None and runNumber > 0:
                        runString = str(runNumber).zfill(9)
                        lfnSuffix = "/%s/%s/%s" % (runString[0:3],
                                                   runString[3:6],
                                                   runString[6:9])
                        unmergedLFN += lfnSuffix
                        mergedLFN += lfnSuffix

                    lfnBase(unmergedLFN)
                    lfnBase(mergedLFN)
                    setattr(outputModule, "processedDataset", processedDataset)

                    # For merge tasks, we want all output to go to the merged LFN base.
                    if taskType == "Merge":
                        setattr(outputModule, "lfnBase", mergedLFN)
                        setattr(outputModule, "mergedLFNBase", mergedLFN)

                        if getattr(outputModule, "dataTier") in ["DQM", "DQMIO"]:
                            datasetName = "/%s/%s/%s" % (getattr(outputModule, "primaryDataset"),
                                                         processedDataset,
                                                         getattr(outputModule, "dataTier"))
                            self.updateDatasetName(datasetName)
                    else:
                        setattr(outputModule, "lfnBase", unmergedLFN)
                        setattr(outputModule, "mergedLFNBase", mergedLFN)

        self.setTaskLogBaseLFN(unmergedLFNBase)

        # do the samething for all the child
        for task in self.childTaskIterator():
            task.updateLFNsAndDatasets(runNumber=runNumber)

        return

    def updateDatasetName(self, datasetName):
        """
        _updateDatasetName_

        Updates the dataset name argument of the mergeTask's harvesting
        children tasks
        """
        for task in self.childTaskIterator():
            if task.taskType() == "Harvesting":
                for stepName in task.listAllStepNames():
                    stepHelper = task.getStepHelper(stepName)

                    if stepHelper.stepType() == "CMSSW":
                        cmsswHelper = stepHelper.getTypeHelper()
                        cmsswHelper.setDatasetName(datasetName)

        return


class WMTask(ConfigSectionTree):
    """
    _WMTask_

    workload management task.
    Allow a set of processing job specifications that are interdependent
    to be modelled as a tree structure.

    """

    def __init__(self, name):
        ConfigSectionTree.__init__(self, name)
        self.objectType = self.__class__.__name__
        self.pathName = None
        self.taskType = None
        self.prepID = None
        self.section_("steps")
        self.steps.topStepName = None
        self.section_("parameters")
        self.section_("pythonLibs")
        self.section_("constraints")
        self.section_("input")
        self.section_("notifications")
        self.section_("subscriptions")
        self.section_("environment")
        self.notifications.targets = []
        self.input.sandbox = None
        self.input.section_("splitting")
        self.input.splitting.algorithm = None
        self.input.splitting.section_("performance")
        self.constraints.section_("sites")
        self.constraints.sites.whitelist = []
        self.constraints.sites.blacklist = []
        self.constraints.sites.trustlists = False
        self.constraints.sites.trustPUlists = False
        self.subscriptions.outputSubs = []
        self.input.section_("WMBS")


def makeWMTask(taskName):
    """
    _makeWMTask_

    Convienience method to instantiate a new WMTask with the name
    provided and wrap it in a helper

    """
    return WMTaskHelper(WMTask(taskName))



"""
_CountFinishedSubscriptionsByTask_

MySQL implementation of Subscription.CountFinishedSubscriptionsByTask
"""

from WMCore.Database.DBFormatter import DBFormatter

class CountFinishedSubscriptionsByTask(DBFormatter):
    """
    Gets count subscriptions given task
    """

    sql = """SELECT ww.name as workflow, ww.task as task, wst.name as jobtype,
             COUNT(case when ws.finished = 1 then 1 else null end) as finished,
             COUNT(case when ws.finished = 0 then 1 else null end) as open,
             COUNT(ws.id) as total,
             MAX(ws.last_update) as updated
             FROM wmbs_subscription ws
             INNER JOIN wmbs_sub_types wst ON wst.id = ws.subtype
             INNER JOIN wmbs_workflow ww ON ww.id = ws.workflow
             GROUP BY ww.name, ww.task, wst.name
          """

    def execute(self, conn = None, transaction = False):
        """
        _execute_

        This DAO returns a list of dictionaries containing
        the key 'id' with the id of the finished subscriptions
        """

        result = self.dbi.processData(self.sql,
                                      conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_GetType_

Oracle implementation of Jobs.GetType
"""




from WMCore.WMBS.MySQL.Jobs.GetType import GetType as MySQLGetType

class GetType(MySQLGetType):
    """
    Identical to MySQL version for now

    """
    pass



#!/usr/bin/env python
"""
_ListDataset_

Oracle implementation of DBSBuffer.GetBlockFromDataset
"""




from WMComponent.DBS3Buffer.MySQL.GetBlockFromDataset import GetBlockFromDataset as MySQLGetBlockFromDataset

class GetBlockFromDataset(MySQLGetBlockFromDataset):
    """
    _GetBlockFromDataset_

    Identical to MySQL version
    """



#!/usr/bin/env python
"""
_CPImpl_

Implementation of StageOutImpl interface for plain cp

Useful for sites with large shared POSIX file systems (HPC)
"""
from __future__ import division

import os
import errno

from WMCore.Storage.Registry import registerStageOutImpl
from WMCore.Storage.StageOutImpl import StageOutImpl


class CPImpl(StageOutImpl):
    """
    _CPImpl_

    """

    def __init__(self, stagein=False):
        StageOutImpl.__init__(self, stagein)
        self.numRetries = 5
        self.retryPause = 300

    def createSourceName(self, protocol, pfn):
        """
        _createSourceName_

        """
        return pfn

    def createOutputDirectory(self, targetPFN):
        """
        _createOutputDirectory_

        create output directory if needed
        """
        targetdir = os.path.dirname(targetPFN)

        if not os.path.isdir(targetdir):
            try:
                os.makedirs(targetdir)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise
        return

    def createStageOutCommand(self, sourcePFN, targetPFN, options=None, checksums=None):
        """
        _createStageOutCommand_

        Build the actual cp stageout command

        """
        copyCommand = ""

        if self.stageIn:
            remotePFN, localPFN = sourcePFN, targetPFN
        else:
            remotePFN, localPFN = targetPFN, sourcePFN
            copyCommand += "LOCAL_SIZE=`stat -c%%s %s`\n" % localPFN
            copyCommand += "echo \"Local File Size is: $LOCAL_SIZE\"\n"

        copyCommand += "cp %s %s\n" % (sourcePFN, targetPFN)

        if self.stageIn:
            copyCommand += "LOCAL_SIZE=`stat -c%%s %s`\n" % localPFN
            copyCommand += "echo \"Local File Size is: $LOCAL_SIZE\"\n"
            removeCommand = ""
        else:
            removeCommand = "rm %s;" % remotePFN

        copyCommand += "REMOTE_SIZE=`stat -c%%s %s`\n" % remotePFN
        copyCommand += "echo \"Remote File Size is: $REMOTE_SIZE\"\n"

        copyCommand += "if [ $REMOTE_SIZE ] && [ $LOCAL_SIZE == $REMOTE_SIZE ]; then exit 0; "
        copyCommand += "else echo \"ERROR: Size Mismatch between local and SE\"; %s exit 60311 ; fi" % removeCommand

        return copyCommand

    def removeFile(self, pfnToRemove):
        """
        _removeFile_

        """
        os.remove(pfnToRemove)
        return

registerStageOutImpl("cp", CPImpl)



#!/usr/bin/env python
"""
_GetUserId_

Oracle implementation of Users.GetUserId
"""

from WMCore.WMBS.MySQL.Users.GetUserId import GetUserId as MySQLGetUserId

class GetUserId(MySQLGetUserId):
    """
    _GetUserId_

    Load user pars by ID
    """
